#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import argparse

import numpy as np

import matplotlib as mpl
mpl.use('Agg') #silent mode
from matplotlib import pyplot as plt
plt.style.use("presentation")

# ASE
import ase.units
from ase import Atoms
from ase.io import read, write
from ase.units import kB
from ase.ga.utilities import (closest_distances_generator, atoms_too_close,
                              atoms_too_close_two_sets)

# calculator
from ase.calculators.gaussian import Gaussian, GaussianOptimizer, GaussianIRC
from ase.calculators.emt import EMT

from xtb.ase.calculator import XTB

# optimisation
from ase.optimize.bfgslinesearch import BFGSLineSearch as QuasiNetwon

# molecular dynamics
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.md.langevin import Langevin

from ase.optimize.minimahopping import PassedMinimum


class AccHopping():

    """
    This method performs accelerated (minima) hopping that automatically 
    finds local minima and transition states under certain thermal conditions. 
    """

    # output files
    traj_name = "miaow.xyz"

    # accelerated MD parameters
    vmax = 0.5 # V_max, eV
    bond_strain_limit = 0.5 # q, maximum bond change compared to the reference state
    P1 = 0.98 # control the curvature near the boundary

    timestep = 1.0 # in fs
    mdstep = 2500


    def __init__(
        self, 
        reactants, 
        temperature,
        mdstep,
        calc, # this is the reference calculator (usually DFT by GAUSSIAN or VASP)
        opt = None, 
        surrogate = None
    ):

        self.beta = 1 / (temperature*kB)

        self.mdstep = mdstep
        self.temperature = temperature

        print("TEMPERATURE: ", self.temperature)
        print("MDSTEP_CYCLE: ", self.mdstep)

        self.calc = calc
        self.opt = opt

        # create initial configuration
        nreactants = len(reactants)
        if nreactants == 1:
            # provide one converged complex as inputs
            self.atoms = reactants[0].copy()
            # TODO: check convergence
            # check number of molecular fragments
            all_tags = set(self.atoms.get_tags())
            print("number of fragments: ", len(all_tags))
        else:
            # generate a random initial structure
            self.atoms = self.__place_random_configuration(reactants)


        # event detection methods
        self._passedminimum = PassedMinimum()

        return

    def __call__(self, step):
        """ run this method
        """

        # check reference state R

        # set MD 
        atoms = self.atoms.copy()
        atoms.calc = self.calc

        self.cur_distance_matrix = self.__calc_bond_list(atoms)
        print("Reference Bond Matrix: ", self.cur_distance_matrix)

        MaxwellBoltzmannDistribution(atoms, temperature_K=self.temperature)
        dyn = Langevin(
            atoms, 
            timestep = self.timestep*ase.units.fs, 
            temperature_K = self.temperature, 
            friction = 0.002, # TODO: what is the unit?
            fixcm = True
        )

        for i in range(step):
            # run hyperdynamics
            self.__molecular_dynamics(dyn, self.mdstep)

            # check if event happens
            self.__detect_event()
            if True:
                self.atoms = None # set current system to quenched state Q
                self.__calc_bond_list()
            else:
                self.atoms = None # restore state to dynamics state D

        return
    
    def __place_random_configuration(self, reactants):
        """ randomly place reactants in a given region
        """
        # tag every molecules and state atomice types
        unique_atom_types = []
        for i, atoms in enumerate(reactants):
            atoms.set_tags(i)
            unique_atom_types.extend(atoms.get_atomic_numbers())

        unique_atom_types = list(set(unique_atom_types))
        blmin = closest_distances_generator(
            atom_numbers=unique_atom_types,
            ratio_of_covalent_radii=0.8 # be careful with test too far
        )
        print(blmin)
        
        # place positions
        nreactants = len(reactants)
        candidate = Atoms("")
        cand_com = np.zeros(3) # set to origin
        for i in range(nreactants):
            atoms = reactants[i].copy()
            for dis in np.arange(2.0, 3.0, 0.05): # distance to the atom group
                print("dis: ", dis)
                # Apply a random translation
                pos = dis*np.linalg.norm(np.random.random(3)) + cand_com
                com = atoms.get_center_of_mass()
                atoms.translate(pos - com) # set to the origin
                # Apply a random rotation to multi-atom blocks
                phi, theta, psi = 360 * np.random.random(3)
                atoms.euler_rotate(
                    phi=phi, theta=0.5 * theta, psi=psi,
                    center=pos
                )
                # check distance
                if not atoms_too_close_two_sets(candidate, atoms, blmin):
                    candidate.extend(atoms)
                    break
            else:
                raise ValueError(f"Failed to insert the {i} reactant in 3.0 AA.")
            cand_com = candidate.get_center_of_mass()
        
        # optimise random formed complex
        self.calc.reset()
        #candidate.calc = self.calc
        #dyn = QuasiNetwon(candidate, trajectory=None)
        #dyn.run(fmax=0.05)

        opt = GaussianOptimizer(candidate, self.calc) # force loose 0.0025 (0.064 eV/AA) tight 0.000015
        opt.run(fmax="loose", steps=100, opt="")

        return candidate
    
    def __calc_bond_list(self, atoms):
        """ calculate bond list for comparision
            Can implement more efficient routine in the future...
        """
        # calc distance matrix
        distance_matrix = atoms.get_all_distances(mic=False, vector=False).copy()
        np.fill_diagonal(distance_matrix, -1.) # self-strain will be None, avoid true-divide in strain

        return distance_matrix
    
    def __molecular_dynamics(self, dyn, maxstep):
        """ perform accelrated dynamics (serial global hyperdynamics)
        """
        st = time.time() # start time

        atoms = dyn.atoms # not copy
        natoms = len(atoms)
        tags = atoms.get_tags() # used to discriminate reactants
        com = atoms.get_center_of_mass().copy()
        print("centre of mass at step 0: ", com)
        dis2com = np.linalg.norm(atoms.positions - com, axis=1)
        print("distances to com: ", dis2com)

        write(self.traj_name, atoms)

        energies = [atoms.get_potential_energy()]
        md_forces = atoms.get_forces().copy() # pristine forces + bias forces

        for i in range(maxstep):
            # update positions with MD
            print(f"\n\n===== MD step: {i} =====")
            dyn.step(md_forces)

            # print MD info
            print("potential energy: ", atoms.get_potential_energy())
            print("temperature: ", atoms.get_temperature())

            # write(self.traj_name, atoms, append=True)
            # continue # test brute-force MD

            # check bond distances
            new_bond_matrix = self.__calc_bond_list(atoms)
            bond_strain = (new_bond_matrix - self.cur_distance_matrix) / self.cur_distance_matrix # not abs

            md_forces = atoms.get_forces().copy()

            # boost_method = "JCP2020"
            boost_method = "JACS2014"
            vboost = 0.
            if boost_method == "JCP2020":
                # ----- JCP2020 serial global method ----- 
                # TODO: create graph and only accelerate atom pairs between molecules
                # also, not consider pairs with too long distance

                # extract from bond list
                max_index = np.unravel_index(np.argmax(np.fabs(bond_strain), axis=None), bond_strain.shape)

                ref_bond_disatnce = self.cur_distance_matrix[max_index]
                new_bond_distance = new_bond_matrix[max_index]
                max_bond_strain = bond_strain[max_index]

                content = "bond pairs: {}-{}\n".format(*max_index)
                content += " current maximum bond strain: {:.4f}\n".format(max_bond_strain)
                content += " distances (ref/new): {:.4f}  {:.4f}\n".format(
                    ref_bond_disatnce, new_bond_distance
                )
                print(content)

                if np.fabs(max_bond_strain) < self.bond_strain_limit:
                    # only apply BB on bond with max strain
                    vboost = self.vmax*(1-(max_bond_strain/self.bond_strain_limit)**2)
                    fboost = 2*self.vmax*max_bond_strain/self.bond_strain_limit**2
            
                    # calc md forces
                    bond_positions = atoms.positions[max_index, :]
                    bias_forces = (
                        fboost * (bond_positions[0, :] - bond_positions[1, :]) 
                        / ref_bond_disatnce / new_bond_distance
                    ) # on atom i
                    print("bias forces on atom i: ", bias_forces)

                    md_forces[max_index, :] = np.vstack([bias_forces, -bias_forces]) 

            elif boost_method == "JACS2014":
                # ----- TODO: JACS2014 adaptive bond boost -----
                # uses pre-defined r_eq not compared to reference bond state
                bmshape = bond_strain.shape[0] # bond matrix shape

                # check bond strain to extract valid bond to boost
                max_index = [0, 0]
                max_bond_strain = 1e-8

                valid_pairs = []
                for i in range(bmshape):
                    for j in range(i+1, bmshape):
                        # check bond strain is smaller than limit, and bond distance is close enough
                        # , and bonds should be formed by different reactants
                        if (
                            np.fabs(bond_strain[i, j]) < self.bond_strain_limit 
                            and new_bond_matrix[i, j] < 3.0 # BB not on too long bonds
                            # and (tags[i] != tags[j])
                        ):
                            valid_pairs.append((i,j))
                            # check max
                            if np.fabs(bond_strain[i, j]) > np.fabs(max_bond_strain):
                                max_bond_strain = bond_strain[i, j]
                                max_index = [i, j]
                nbonds = len(valid_pairs)
                if nbonds > 0:
                    print("number of bonds: ", nbonds)
                    print(valid_pairs)
                else:
                    print("no valid bond pairs, skip this step...")
                    continue

                max_index = tuple(max_index)
                ref_bond_disatnce = self.cur_distance_matrix[max_index]
                new_bond_distance = new_bond_matrix[max_index]

                content = "bond pairs: {}-{}\n".format(*max_index)
                content += " current maximum bond strain: {:.4f}\n".format(max_bond_strain)
                content += " distances (ref/new): {:.4f}  {:.4f}\n".format(
                    ref_bond_disatnce, new_bond_distance
                )
                print(content)

                # calculate boost energy
                vboost_separate = self.vmax / nbonds * (1 - (bond_strain / self.bond_strain_limit)**2)
                vboost = 0.
                for (i, j) in valid_pairs:
                    vboost += vboost_separate[i, j]

                fboost = 2*self.vmax/nbonds*bond_strain/self.bond_strain_limit**2

                bond_ratio_square = (max_bond_strain/self.bond_strain_limit)**2
                envelope = (1-bond_ratio_square)*((1-bond_ratio_square)/(1-self.P1**2*bond_ratio_square))

                # calculate bias forces
                for (i, j) in valid_pairs:
                    # use new distance
                    ref_dis = self.cur_distance_matrix[i, j]
                    new_dis = new_bond_matrix[i, j]
                    bond_positions = atoms.positions[(i,j), :]
                    bias_forces = (
                        envelope * fboost[i, j] * (bond_positions[0, :] - bond_positions[1, :])
                        / ref_dis / new_dis
                    )
                    md_forces[(i,j), :] = np.vstack([bias_forces, -bias_forces]) 

                # add extra for max_index
                uuu = max_bond_strain / self.bond_strain_limit
                xxx = 1 - uuu**2
                mmm = 1 - self.P1**2*uuu**2
                bond_positions = atoms.positions[max_index, :]
                envelope_grad = (
                    (-2*uuu * xxx / mmm) + (xxx * (-2*uuu*mmm - xxx*(-2*self.P1**2*uuu)) / mmm**2)
                ) * (
                    (bond_positions[0, :] - bond_positions[1, :]) / ref_bond_disatnce / new_bond_distance / self.bond_strain_limit
                )
                bias_forces = -envelope_grad * vboost
                md_forces[max_index, :] = np.vstack([bias_forces, -bias_forces])

            else:
                raise ValueError(f"Unknown boost method {boost_method}...")
            
            # TODO: add wall potential to avoid molecules moving away
            # harmonic V=1/2k(x-c)^2 F = -k(x-c)*dx/dr
            centre_of_spring = 3.6
            spring_constant = 2000.
            positions = atoms.positions # current positions
            for i in range(natoms):
                dvec = positions[i, :] - com
                dis = np.linalg.norm(dvec)
                if dis > centre_of_spring:
                    wall_force = -spring_constant*np.fabs(dis-centre_of_spring)*dvec/dis
                    md_forces[i, :] += wall_force

            # calculate boost factor
            boost_factor = np.exp(self.beta*vboost)
            print("boost_factor: {:.4f}".format(boost_factor))

            # save trajectory
            write(self.traj_name, atoms, append=True)

            # TODO: check event
            energies.append(atoms.get_potential_energy())
            passedmin = self._passedminimum(energies)
            if passedmin:
                print("Found Minimum!!!")
            else:
                print("NO MINIMUM...")
        
        et = time.time()

        print("Time for amd: {:.4f} s".format(et-st))

        return
    
    def __detect_event(self):
        """ Check a reactive event has passed using various methods
            1. Use minima hopping method JCP2004 similar to ASE one
            2. TODO: JCTC2016
            3. max atomic displacement usually for solid-state systems
        """

        return
    
    def __optimisation(self):
        """ perform optimisation on current structure
            quech the stucture to the nearby local minimum
        """

        return
    
    def __place_bias_on_minima(self):
        """ place bias on given minima
        """

        return


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-t", "--temperature", default=673, type=float,
        help="simulation temperature"
    )
    parser.add_argument(
        "-n", "--mdstep", default=100, type=int,
        help="number of mdsteps"
    )

    args = parser.parse_args()

    # load reactants
    reactants = []
    atoms = read("/mnt/scratch2/users/40247882/amdrs/opt/H2CO/H2CO_opt.xyz")
    reactants.append(atoms.copy())
    atoms = read("/mnt/scratch2/users/40247882/amdrs/opt/H2CCHOH/H2CCHOH_opt.xyz")
    reactants.append(atoms.copy())

    # load optimised initial structure
    reactants = []
    atoms = read("/mnt/scratch2/users/40247882/amdrs/rs/test-structures/IS-1_opt.xyz")
    reactants.append(atoms.copy())

    # prepare calculator
    calc = Gaussian(
        directory="Gau-Worker",
        label="opt",
        chk = "MyJob.chk",
        mem = "1GB",
        nprocshared = 4,
        method = "b3lyp",
        # basis = "6-31G",
        basis = "6-31G"
    )
    # calc = EMT()
    # calc = XTB(method="GFN2-xTB")

    rs = AccHopping(reactants, args.temperature, args.mdstep, calc)
    rs(10)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
from typing import Optional, Tuple, List

import numpy as np

from ase import Atoms
from ase.calculators.calculator import Calculator
from ase.geometry import find_mic

from ..timeio import TimeIOCalculator


def compute_distance(cell, positions, pbc: bool = True):
    """"""
    # compute colvar
    assert positions.shape[0] == 2
    vec, dis = find_mic(positions[0] - positions[1], cell=cell)

    return vec, dis


def compute_distance_harmonic_energy_and_forces(
    vec, dis: float, center: float, kspring: float
):
    """"""
    # compute energy
    energy = np.sum(0.5 * kspring * (dis - center) ** 2)

    # compute forces
    forces = np.zeros((2, 3))
    frc_i = -kspring * vec / dis
    forces[0] += frc_i
    forces[1] += -frc_i

    return energy, forces


class DistanceHarmonicCalculator(TimeIOCalculator):

    implemented_properties = ["energy", "free_energy", "forces"]

    def __init__(
        self, group: List[int], center: float, kspring: float = 0.1, *args, **kwargs
    ):
        """"""
        super().__init__(*args, **kwargs)

        num_group_atoms = len(group)
        assert num_group_atoms == 2
        self.group = group

        self.center = center

        self.kspring = kspring

        return

    def _icalculate(self, atoms, properties, system_changes) -> Tuple[dict, list]:
        """"""
        vec, dis = compute_distance(atoms.cell, atoms.positions[self.group], pbc=True)

        energy, ext_forces = compute_distance_harmonic_energy_and_forces(
            vec, dis, self.center, self.kspring
        )
        forces = np.zeros(atoms.positions.shape)
        forces[self.group] = ext_forces

        results = {}
        results["energy"] = energy
        results["free_energy"] = energy
        results["forces"] = forces

        step_info = (self.num_steps, dis, energy)

        return results, step_info

    def _write_first_step(self):
        """"""
        content = f"# {self.pace =} {self.group =} {self.center =} {self.kspring =}\n"
        content += "# {:>10s}  {:>12s}  {:>12s}\n".format("step", "distance", "energy")
        with open(self.log_fpath, "w") as fopen:
            fopen.write(content)

        return
    
    def _write_step(self):
        """"""
        content = "{:>12d}  {:>12.4f}  {:>12.4f}\n".format(*self.step_info)
        with open(self.log_fpath, "a") as fopen:
            fopen.write(content)

        return

    @staticmethod
    def broadcast_params(inp_dict: dict) -> List[dict]:
        """"""
        # broadcast center or kspring
        centers = inp_dict.get("center", [])
        if isinstance(centers, float):
            centers = [centers]
        num_centers = len(centers)

        ksprings = inp_dict.get("kspring", [])
        if isinstance(ksprings, float):
            ksprings = [ksprings]
        num_ksprings = len(ksprings)

        new_inputs = []
        if num_centers == 1 and num_ksprings == 1:
            new_inputs = [inp_dict]
        elif num_centers == 1 and num_ksprings > 1:
            new_inputs = [copy.deepcopy(inp_dict) for _ in range(num_ksprings)]
            for i, x in enumerate(new_inputs):
                x["kspring"] = ksprings[i]
        elif num_centers > 1 and num_ksprings == 1:
            new_inputs = [copy.deepcopy(inp_dict) for _ in range(num_centers)]
            for i, x in enumerate(new_inputs):
                x["center"] = centers[i]
        else:
            raise RuntimeError("Broadcast cannot.")

        return new_inputs
    
    @staticmethod
    def broadcast(inp_dict: dict) -> List["DistanceHarmonicCalculator"]:
        """Create a list of calculators based on input parameters."""

        new_inputs = DistanceHarmonicCalculator.broadcast_params(inp_dict)
        calcs = [DistanceHarmonicCalculator(**x) for x in new_inputs]

        return calcs


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*


import warnings

from .. import registers
from ..manager import AbstractPotentialManager
from ..trainer import AbstractTrainer

from ..calculators.dummy import DummyCalculator
from ..calculators.mixer import CommitteeCalculator


# - basic potentials
# -- MLIP
from .deepmd import DeepmdManager, DeepmdTrainer, DeepmdDataloader
registers.manager.register(DeepmdManager)
registers.trainer.register(DeepmdTrainer)
registers.dataloader.register(DeepmdDataloader)

from .reann.beann import BeannManager, BeannTrainer
registers.manager.register(BeannManager)
registers.trainer.register(BeannTrainer)

from .reann.reann import ReannManager, ReannTrainer, ReannDataloader
registers.manager.register(ReannManager)
registers.trainer.register(ReannTrainer)
registers.dataloader.register(ReannDataloader)

from .lasp import LaspManager
registers.manager.register(LaspManager)

from .mace import MaceManager, MaceTrainer, MaceDataloader
registers.manager.register(MaceManager)
registers.trainer.register(MaceTrainer)
registers.dataloader.register(MaceDataloader)

from .nequip import NequipManager, NequipTrainer
registers.manager.register(NequipManager)
registers.trainer.register(NequipTrainer)

#try:
#    from .schnet import SchnetManager
#    registers.manager.register(SchnetManager)
#except ImportError as e:
#    warnings.warn("Module {} import failed: {}".format("schnet", e), UserWarning)

# -- reference potentials
# --- DFT
from .cp2k import Cp2kManager
registers.manager.register(Cp2kManager)

from .espresso import EspressoManager
registers.manager.register(EspressoManager)

from .vasp import VaspManager
registers.manager.register(VaspManager)

# --- FFs
from .asepot import AsePotManager
registers.manager.register("AseManager")(AsePotManager)

from .eam import EamManager
registers.manager.register(EamManager)

from .emt import EmtManager
registers.manager.register(EmtManager)

from .reax import ReaxManager
registers.manager.register(ReaxManager)

# - advanced potentials
from .grid import GridManager
registers.manager.register(GridManager)

from .mixer import MixerManager
registers.manager.register(MixerManager)

# - optional potentials
try:
    from .xtb import XtbManager
    registers.manager.register(XtbManager)
except ImportError as e:
    warnings.warn("Module {} import failed: {}".format("xtb", e), UserWarning)

try:
    from .dftd3 import Dftd3Manager
    registers.manager.register(Dftd3Manager)
except ImportError as e:
    warnings.warn("Module {} import failed: {}".format("dftd3", e), UserWarning)

try:
    from .bias import BiasManager
    registers.manager.register(BiasManager)
except ImportError as e:
    warnings.warn("Module {} import failed: {}".format("bias", e), UserWarning)

from .plumed.plumed import PlumedManager
registers.manager.register(PlumedManager)

# - trainers
from .gp.fgp import FGPTrainer
registers.trainer.register("FgpTrainer")(FGPTrainer)

from .gp.sgp import SGPTrainer
registers.trainer.register("SgpTrainer")(SGPTrainer)



if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from pathlib import Path
from typing import Union, List, NoReturn

import numpy as np

from ase import Atoms
from ase.io import read, write

from ..data.array import AtomsNDArray
from .selector import AbstractSelector


class ComposedSelector(AbstractSelector):
    """Perform several selections consecutively."""

    name = "composed"

    default_parameters = dict(selectors=[])

    def __init__(
        self, selectors: List[AbstractSelector], directory="./", *args, **kwargs
    ):
        """"""
        super().__init__(directory=directory, *args, **kwargs)

        self.selectors = selectors

        return

    def _mark_structures(self, frames: AtomsNDArray, *args, **kwargs) -> None:
        """Return selected indices."""
        # - update selectors' directories
        for s in self.selectors:
            s.directory = self._directory

        # - initial index stuff
        curr_frames = frames

        # - run selection
        for i, node in enumerate(self.selectors):
            # - adjust name
            prev_fname = node._fname
            node.fname = str(i) + "-" + prev_fname
            # - map indices
            #   TODO: use _select_indices instead?
            node.select(curr_frames)

            node.fname = prev_fname

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy

from typing import List, Callable

import numpy as np

from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    ...

from ase.formula import Formula
from ase.geometry import find_mic
from ase.neb import NEB, NEBTools


def convert_index_to_formula(atoms, group_indices: List[List[int]]):
    """"""
    formulae = []
    for g in group_indices:
        symbols = [atoms[i].symbol for i in g]
        formulae.append(
            Formula.from_list(symbols).format("hill")
        )
    #formulae = sorted(formulae)

    return formulae

def compute_rxn_coords(frames):
    """Compute reaction coordinates."""
    # - avoid change atoms positions and lost energy properties...
    nframes = len(frames)
    natoms = len(frames[0])
    coordinates = np.zeros((nframes, natoms, 3))
    for i, a in enumerate(frames):
        coordinates[i, :, :] = copy.deepcopy(frames[i].get_positions())

    rxn_coords = []
    cell = frames[0].get_cell(complete=True)
    for i in range(1, nframes):
        prev_positions = coordinates[i-1]
        curr_positions = coordinates[i]
        shift = curr_positions - prev_positions
        curr_vectors, curr_distances = find_mic(shift, cell, pbc=True)
        coordinates[i] = prev_positions + curr_vectors
        rxn_coords.append(np.linalg.norm(curr_vectors))

    rxn_coords = np.cumsum(rxn_coords)
    rxn_coords = np.hstack(([0.], rxn_coords))

    return rxn_coords

def plot_mep(wdir, images):
    """"""
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12, 8))
    plt.suptitle("Nudged Elastic Band Calculation")

    nbt = NEBTools(images=images)
    nbt.plot_band(ax=ax)

    plt.savefig(wdir/"neb.png")

    plt.close()

    return

def plot_bands(wdir, images, nimages: int):
    """"""
    #print([a.get_potential_energy() for a in images])
    
    nframes = len(images)

    nbands = int(nframes/nimages)

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12, 8))
    plt.suptitle("Nudged Elastic Band Calculation")

    for i in range(nbands):
        #print(f"plot_bands {i}")
        nbt = NEBTools(images=images[i*nimages:(i+1)*nimages])
        nbt.plot_band(ax=ax)

    plt.savefig(wdir/"bands.png")

    return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from inspect import iscoroutine
import numpy as np

from ase.io import read, write
from ase.constraints import FixAtoms, FixBondLengths
from ase.calculators.emt import EMT
from ase.neb import NEB
#from ase.neb import SingleCalculatorNEB
from ase.optimize import BFGS
from ase.geometry import wrap_positions
from sklearn.preprocessing import scale
from torch import saddmm


class AccNEB():

    def __init__(self):

        return
    
    def __call__(self):

        return


def test_cons(atoms, calc):
    atoms.calc = calc
    atoms.set_constraint(
        FixBondLengths(
            pairs = [[0, 23]],
            bondlengths = [2.426]
        )
    )

    qn = BFGS(atoms, trajectory="cons.traj")
    qn.run(fmax=0.05, steps=50)

    return



if __name__ == "__main__":
    # test LASP-NEB
    from gdpx.computation.lasp import LaspNN
    pot_path = "/mnt/scratch2/users/40247882/catsign/lasp-main/ZnCrOCH.pot"
    pot = dict(
        C  = pot_path,
        O  = pot_path,
        Cr = pot_path,
        Zn = pot_path
    )
    calc = LaspNN(
        directory = "./LaspNN-Worker",
        command = "mpirun -n 4 lasp",
        pot=pot
    )


    initial = read("/mnt/scratch2/users/40247882/catsign/ZnOx/P188/CO-IS.xyz")
    transition = read("/mnt/scratch2/users/40247882/catsign/ZnOx/P188/CO-TS.xyz")
    final = read("/mnt/scratch2/users/40247882/catsign/ZnOx/P188/CO-FS.xyz")

    transition = read("/mnt/scratch2/users/40247882/catsign/ZnOx/cases/TStest.xyz")

    cons_inddices = list(range(1,13)) + list(range(24,36))
    constraint = FixAtoms(indices=cons_inddices) # first atom is O
    #constraint = FixAtoms(indices=[1,2,3,4,5,6,7,8]) # first atom is O

    transition.set_constraint(constraint)
    transition.calc = calc

    # test_cons(transition, calc)
    constrdyn = AccCons(transition, cpairs=[0,23])
    constrdyn()
    exit()

    nimages = 7
    images = [initial]
    images += [initial.copy() for i in range(nimages-2)]
    images.append(final)

    for atoms in images:
        calc = LaspNN(
            directory = "./LaspNN-Worker",
            command = "mpirun -n 4 lasp",
            pot=pot
        )
        atoms.calc = calc
        atoms.set_constraint(constraint)

    neb = NEB(
        images, 
        allow_shared_calculator=False,
        k = 0.1, # spring constant
        climb = False,
        # dynamic_relaxation = False
    )
    #neb = SingleCalculatorNEB(images)

    neb.interpolate()
    #print(neb.images)

    qn = BFGS(neb, trajectory="neb.traj")
    qn.run(fmax=0.05, steps=50)

    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import dataclasses
import pathlib
from typing import Callable, Optional, List, Mapping
from itertools import chain

import numpy as np
import networkx as nx

from ase import Atoms
from ase.io import read, write
from ase.neighborlist import natural_cutoffs, NeighborList

from .. import config

from .utils import grid_iterator, node_symbol, bond_symbol, unpack_node_name

"""Create the graph representation of a structure based on its neighbour list.
"""

DEFAULT_NEIGH_PARAMS = dict(
    covalent_ratio = 1.0,
    skin = 0.0,
    self_interaction = False,
    bothways = True,
)

DIS_SURF2SURF = 2
DIS_ADS2SURF = 1

@dataclasses.dataclass
class GraphRadius:

    """Define some graph edge attributes."""

    radius: int = 2


class NeighGraphCreator():

    # Parameters for ASE neighbor list.
    covalent_ratio: float= 1.0 #: Multiplier for the covalent bond distance.
    skin = 0.0 # TODO: the neighbours will be in cutoff + skin
    self_interaction = False #: Whether consider self-interaction under PBC.
    bothways = True #: Whether return the full neighbor list.

    #: PBC grid.
    #pbc_grid: List[int] = [2, 2, 0] # z=0 is for surface systems.

    def __init__(
        self, covalent_ratio=1.0, skin=0.0, self_interaction=False, bothways=True, 
        *args, **kwargs
    ):
        """"""
        #for key, value in kwargs.items():
        #    if hasattr(self, key):
        #        setattr(self, key, value)
        self.covalent_ratio = covalent_ratio
        self.skin = skin
        self.self_interaction = self_interaction
        self.bothways = bothways

        return
    
    def build_neighlist(self, atoms, bothways=True):
        # NOTE: for debug only
        #content = "----- Build NeighborList -----\n"
        #content += ("mult: {} skin: {} bothways: {}\n").format(
        #    self.covalent_ratio, self.skin, bothways
        #)
        #print(content)

        nl = NeighborList(
            natural_cutoffs(atoms, mult=self.covalent_ratio), 
            skin = self.skin, sorted=False,
            self_interaction=self.self_interaction, 
            bothways = bothways
        )
        #print(natural_cutoffs(atoms, mult=self.covalent_ratio))

        return nl


class StruGraphCreator():

    """Create the graph representation of a structure based on its neighbour list.
    """
    #: Distance [Ang].
    substrate_adsorbate_distance: int = 2.5 

    #: Working directory.
    _directory = "./"

    #: Standard print function.
    _print: Callable = config._print

    #: Standard debug function.
    _debug: Callable = config._debug

    def __init__(
        self, adsorbate_elements: List[str],
        pbc_grid: List[int]=[2,2,0], graph_radius: int=2, 
        neigh_params: dict=DEFAULT_NEIGH_PARAMS, 
        *args, **kwargs
    ):
        """Init graph creator.

        Args:
            adsorbate_elements: Elements that may be considered as adsorbates.
            graph_radius: 
                The subgraph size for comparasion. 
                # help='Sets the graph radius, this can be tuned for different behavior')

        """
        # NOTE: half neigh is enough for creating structure graph
        self.adsorbate_elements = adsorbate_elements
        # self.adsorbate_indices = adsorbate_indices

        self.graph_radius = graph_radius
        self.pbc_grid = pbc_grid

        # - neigh params
        # covalent_ratio = kwargs.pop("covalent_ratio", 1.0)
        # pbc_grid = kwargs.pop("pbc_grid", [2,2,0])
        # skin = kwargs.pop("skin", 0.0)
        # self_interaction = kwargs.pop("self_interaction", False)
        # bothways = kwargs.pop("bothways", True)
        # self.neigh_creator = NeighGraphCreator(
        #    covalent_ratio=covalent_ratio, skin=skin,
        #    self_interaction=self_interaction, bothways=bothways,
        #    pbc_grid=pbc_grid, *args, **kwargs
        # )
        self.neigh_creator = NeighGraphCreator(**neigh_params)

        return

    @property
    def directory(self):

        return self._directory

    @directory.setter
    def directory(self, directory_):
        """"""
        self._directory = pathlib.Path(directory_)

        return

    @property
    def DIS_SURF2SURF(self):

        return 2

    @property
    def DIS_ADS2SURF(self):

        return 1

    def add_atoms_node(self, graph, sym1, a1, o1, **kwargs):
        """Add an atom node to the graph.

        Args:
            graph: Graph.
            a1: Atom index.
            o1: Atom offset in PBC.

        """
        graph.add_node(
            node_symbol(sym1, a1, o1), index=a1, central_ads=False, **kwargs
        )

        return

    def add_atoms_edge(
        self, 
        graph,
        sym1, sym2,
        a1, a2, o1, o2, 
        dis,
        adsorbate_atoms, 
        **kwargs
    ):
        """ graph edge format
            dist 
                2 bond in substrate
                1 bond between adsorbate and substrate
                0 bond in adsorbate
            ads_only 
                0 means the bond is in the adsorbate
                2 otherwise
        """

        dist = self.DIS_SURF2SURF - (self.DIS_ADS2SURF if a1 in adsorbate_atoms else 0) - (self.DIS_ADS2SURF if a2 in adsorbate_atoms else 0)

        graph.add_edge(
            # node index
            node_symbol(sym1, a1, o1),
            node_symbol(sym2, a2, o2),
            # attributes - edge data
            bond = bond_symbol(sym1, sym2, a1, a2),
            index = "{}:{}".format(*sorted([a1, a2])),
            dist = dist,
            dist_edge = dis,
            ads_only = 0 if (a1 in adsorbate_atoms and a2 in adsorbate_atoms) else 2,
            **kwargs
        )

        return

    def check_system(
        self, atoms: Atoms, min_vacuum_ratio: float=0.3,
        substrate_indices: Optional[List[int]] = None
    ) -> str:
        """Determine system type.

        Args:
            atoms: Atoms object.
            substrate_indices: Atomic indices used to determine the system type.

        Return:
            A string that desribes the system type.

        """
        # TODO: solvent-solid interface

        cell_length = np.linalg.norm(atoms.get_cell(), axis=1)
        coord_range = np.max(atoms.positions, axis=0) - np.min(atoms.positions, axis=0)

        vacuum_ratio = (cell_length - coord_range) / cell_length
        num_vacuum = sum(vacuum_ratio >= min_vacuum_ratio)
        self._print(f"{vacuum_ratio = }")

        if num_vacuum == 3:
            system = "cluster"  # or molecule
        elif num_vacuum == 2:  # nanowire?
            # TODO: find direction?
            system = "wire"
        elif num_vacuum == 1:
            # TODO: find direction?
            system = "surface"
        else:  # num_vacuum == 0
            system = "bulk"

        return system

    def generate_graph(
        self, atoms: Atoms, ads_indices: List[int]=None, 
        clean_substrate: Atoms=None
    ) -> nx.Graph:
        """Generate a graph for the input atoms."""
        # - create graph with adsorbates
        full = self._generate_structure_graph(atoms, ads_indices)
        input_ads_indices = copy.deepcopy(ads_indices)

        # - create a graph of clean substrate plus adsorbate graph
        #   to avoid surface relaxation
        if clean_substrate is not None: # BUG: need an alignment operation
            self._print("use clean substrate plus adsorbates...")
            clean_graph = self._generate_structure_graph(clean_substrate)
            # Read all the edges, that are between adsorbate and surface (dist<2 condition)
            ads_edges = [(u, v, d) for u, v, d in full.edges.data() if d["dist"] < self.DIS_SURF2SURF]
            # take all the nodes that have an adsorbate atoms
            ads_nodes = [(n, d) for n, d in full.nodes.data() if d["index"] in input_ads_indices]
            # for (n, d) in ads_nodes:
            #    print(n, d)
            full = nx.Graph(clean_graph)
            full.add_nodes_from(ads_nodes)
            full.add_edges_from(ads_edges)
        else:
            ...

        return full

    def _generate_structure_graph(self, atoms_, ads_indices: List[int]=None):
        """ generate molecular graph for reaction detection
            this is part of process_atoms
        """
        # NOTE: create a new graph when this method is called
        graph = nx.Graph() 

        atoms = atoms_

        # init few params
        grid = self.pbc_grid

        # - add all atoms to graph
        natoms = len(atoms)
        for i in range(natoms):
            for x, y, z in grid_iterator(grid):
                self.add_atoms_node(graph, atoms[i].symbol, i, (x, y, z))   

        # - create a neighbour list
        nl = self.neigh_creator.build_neighlist(atoms, bothways=False)
        nl.update(atoms)

        # - add all edges to graph
        for centre_idx in range(natoms):
            for x, y, z in grid_iterator(grid):
                nei_indices, offsets = nl.get_neighbors(centre_idx)
                for nei_idx, offset in zip(nei_indices, offsets):
                    ox, oy, oz = offset
                    if not (-grid[0] <= ox + x <= grid[0]):
                        continue
                    if not (-grid[1] <= oy + y <= grid[1]):
                        continue
                    if not (-grid[2] <= oz + z <= grid[2]):
                        continue
                    # TODO: use tag to manually set dist between atoms in one adsorbate
                    # This line ensures that only surface-adsorbate bonds are accounted for that are less than 2.5 Å
                    dis = atoms.get_distances(centre_idx, nei_idx, mic=True)
                    if dis > self.substrate_adsorbate_distance and (bool(centre_idx in ads_indices) ^ bool(nei_idx in ads_indices)):
                        continue
                    centre_sym, nei_sym = atoms[centre_idx].symbol, atoms[nei_idx].symbol
                    self.add_atoms_edge(
                        graph, centre_sym, nei_sym, centre_idx, nei_idx, (x, y, z), (x + ox, y + oy, z + oz), 
                        atoms.get_distance(centre_idx, nei_idx, mic=True),
                        ads_indices
                    )

        return graph

def extract_chem_envs(
    graph: nx.Graph, atoms: Atoms, ads_indices: Optional[List[int]],
    graph_radius: int=2,
):
    """Extract chemical environments of selected species.
    
    Part of process_atoms. The species could either a single atom or a molecule.
    
    Args:
        graph: Atoms graph.
        atoms: Atoms object.
        ads_indices: A List of indices.
        graph_radius: The graph scope that defines a local environment.

    Returns:
        Chemical environments of species.

    """
    # 
    full = graph

    # - Get adsorbate graphs (single atom or molecule)
    # All adsorbates into single graph, no surface
    ads_nodes = None
    ads_nodes = [node_symbol(atoms[i].symbol, i, (0, 0, 0)) for i in ads_indices]
    ads_graphs = nx.subgraph(full, ads_nodes)

    # - get subgraphs, one for each molecule
    #ads_graphs = nx.connected_component_subgraphs(ads_graphs) # removed in v2.4
    # this creates a list of separate adsorbate graphs
    ads_graphs = [ads_graphs.subgraph(c) for c in nx.connected_components(ads_graphs)]
    #print("number of adsorbate graphs: ", len(ads_graphs))

    chem_envs = []
    for idx, ads in enumerate(ads_graphs):
        #print(f"----- adsorbate {idx} -----")
        #plot_graph(ads, fig_name=f"ads-{idx}.png")
        #print("ads nodes: ", ads.nodes())
        initial = list(ads.nodes())[0] # the first node in the adsorbate
        full_ads = nx.ego_graph(full, initial, radius=0, distance="ads_only") # all nodes in this adsorbate, equal ads?
        #print("full ads: ", full_ads.nodes())

        new_ads = nx.ego_graph(
            full, initial, 
            radius=(graph_radius*DIS_SURF2SURF)+DIS_ADS2SURF, 
            distance="dist"
        ) # consider neighbour atoms
        new_ads = nx.Graph(nx.subgraph(full, list(new_ads.nodes()))) # return a new copy of graph

        # update attribute of this adsorbate
        for node in ads.nodes():
            new_ads.add_node(node, central_ads=True) # means the atom is in [0,0,0]

        # update attr of this and neighbour adsorbates
        for node in full_ads.nodes():
            new_ads.add_node(node, ads=True)
        #print("new ads: ", new_ads.nodes())
        
        #plot_graph(new_ads, fig_name=f"ads-{idx}.png")

        chem_envs.append(new_ads)
    #exit()
    
    # - unique and sort
    #chem_envs = self.unique_adsorbates(chem_envs)  
    ## sort chem_env by number of edges
    #chem_envs.sort(key=lambda x: len(x.edges()))

    return chem_envs.copy()


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*


import copy
from typing import List

from . import registers
from . import AbstractPotentialManager, DummyCalculator
from ..calculators.mixer import EnhancedCalculator

from ase.calculators.calculator import Calculator


class MixerManager(AbstractPotentialManager):

    name = "mixer"
    implemented_backends = ("ase",)

    valid_combinations = (
        # calculator, dynamics
        ("ase", "ase"),
        ("ase", "lammps"),
    )

    def __init__(self) -> None:
        """"""

        return

    def register_calculator(self, calc_params) -> None:
        """"""
        super().register_calculator(calc_params)

        # some basic parameters
        save_host = calc_params.get("save_host", True)

        # process potters
        potters_ = calc_params.get("potters", [])
        npotters = len(potters_)
        assert npotters > 1, "Mixer needs at least two potters."

        potters = []
        for potter_ in potters_:
            if isinstance(potter_, AbstractPotentialManager):
                potter = potter_
            else:  # assume it is a dict
                name = potter_.get("name", None)
                potter = registers.create("manager", name, convert_name=True)
                potter.register_calculator(potter_.get("params", {}))
            potters.append(potter)

        # Used by ASE to determine timestep, dump_period ...
        self.potters = potters

        # try broadcasting calculators
        broadcast_index = -1
        for i, p in enumerate(potters):
            if isinstance(p.calc, Calculator):
                ...
            else:  # assume it is a List of calculators
                if broadcast_index != -1:
                    raise RuntimeError(
                        f"Broadcast cannot on {broadcast_index} and {i}."
                    )
                broadcast_index = i

        calc = DummyCalculator()
        if self.calc_backend == "ase":
            if broadcast_index == -1:
                pot_calcs = [p.calc for p in potters]
                calc = EnhancedCalculator(pot_calcs, save_host=save_host)
            else:
                # non-broadcasted calculators are shared...
                num_instances = len(potters[broadcast_index].calc)
                new_pot_calcs = []
                for i in range(num_instances):
                    x = [p.calc for p in potters]
                    x[broadcast_index] = potters[broadcast_index].calc[i]
                    new_pot_calcs.append(x)
                calc = [
                    EnhancedCalculator(x, save_host=save_host) for x in new_pot_calcs
                ]
        else:
            ...

        self.calc = calc

        return

    def switch_uncertainty_estimation(self, status: bool=True):
        """"""
        if self.calc_backend == "ase":
            has_switched = False
            for manager in self.potters:
                if hasattr(manager, "switch_uncertainty_estimation"):
                    manager.switch_uncertainty_estimation(status=status)
                    has_switched = True
                else:
                    ...
            if has_switched:
                self.calc_params["potters"] = self.potters
                self.register_calculator(self.calc_params)
        else:
            ...

        return

    def as_dict(self) -> dict:
        """"""
        params = super().as_dict()
        params["params"]["potters"] = [p.as_dict() for p in self.potters]

        return params

    @staticmethod
    def broadcast(manager: "MixerManager") -> List["MixerManager"]:
        """"""
        calc_params = copy.deepcopy(manager.calc_params)
        calc_params["backend"] = manager.calc_backend
        # print(f"{calc_params =}")

        # HACK: Here we use potter.calc to dertermine whether
        #       the potter is broadcastable
        broadcast_index = -1
        for i, p in enumerate(manager.potters):
            if isinstance(p.calc, Calculator):
                ...
            else:  # assume it is a List of calculators
                if broadcast_index != -1:
                    raise RuntimeError(
                        f"Broadcast cannot on {broadcast_index} and {i}."
                    )
                broadcast_index = i

        if broadcast_index != -1:
            b_potter = manager.potters[broadcast_index]
            b_potter_cls = b_potter.__class__
            # print(f"{b_potter_cls =}")
            b_potters = b_potter_cls.broadcast(manager=b_potter)
            # print(f"{b_potters =}")
            num_potters = len(b_potters)
            broadcasted_params = []
            for i in range(num_potters):
                x = copy.deepcopy(calc_params)
                x["potters"][broadcast_index] = b_potters[i]
                broadcasted_params.append(x)
            managers = []
            for inp_dict in broadcasted_params:
                manager = MixerManager()
                manager.register_calculator(inp_dict)
                managers.append(manager)
        else:
            managers = [manager]

        return managers


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import io
import gzip
import pathlib
import time
import uuid
import shutil
import tarfile
import yaml

from typing import Optional, Union, List

from tinydb import Query, TinyDB

import numpy as np

from ase import Atoms
from ase.io import read, write

from .worker import AbstractWorker
from ..utils.command import CustomTimer


class SingleWorker(AbstractWorker):

    #: Prefix of the computation folder.
    COMP_PREFIX: str = "cand"

    #: TODO: Current working directory name...
    _wdir_name: Optional[Union[str, pathlib.Path]] = None

    #: How to retrieve computation results, which should be `single` or `all`.
    _retrieve_mode: str = "single"

    def __init__(self, potter, driver, scheduler, directory="./", *args, **kwargs) -> None:
        """"""
        super().__init__(directory)

        self.potter = potter
        self.driver = driver
        self.scheduler = scheduler

        return
    
    @staticmethod
    def from_a_worker(worker) -> "SingleWorker":
        """"""
        single_worker = SingleWorker(
            worker.potter, worker.driver, worker.scheduler,
            worker.directory
        )

        return single_worker
    
    @property
    def wdir_name(self):
        """"""

        return self._wdir_name
    
    @wdir_name.setter
    def wdir_name(self, name: str):
        """"""
        self._wdir_name = name

        return 
    
    def run(self, builder, *args, **kwargs):
        """This worker accepts only a single structure."""
        super().run(*args, **kwargs)

        if isinstance(builder, list): # assume List[Atoms]
            frames = builder
        else: # assume it is a builder
            frames = builder.run()
        nframes = len(frames)
        assert len(frames) == 1, f"{self.__class__.__name__} accepts only a single structure."

        uid = str(uuid.uuid1())
        assert self.wdir_name is not None, "Computation folder is not set."
        wdir = self.directory / self.wdir_name
        job_name = uid + "-" + "single"

        scheduler = self.scheduler
        if scheduler.name == "local":
            with CustomTimer(name="run-driver", func=self._print):
                self.driver.directory = wdir
                self._print(
                    f"{time.asctime( time.localtime(time.time()) )} {wdir.name} {self.driver.directory.name} is running..."
                )
                self.driver.reset()
                self.driver.run(frames[0], read_ckpt=True, extra_info=None)
        else:
            worker_params = {}
            worker_params["use_single"] = True
            worker_params["driver"] = self.driver.as_dict()
            worker_params["potential"] = self.potter.as_dict()

            with open(wdir/f"worker-{uid}.yaml", "w") as fopen:
                yaml.dump(worker_params, fopen)

            # - save structures
            dataset_path = str((wdir/f"_gdp_inp.xyz").resolve())
            write(dataset_path, frames[0])

            # - save scheduler file
            jobscript_fname = f"run-{uid}.script"
            self.scheduler.job_name = job_name
            self.scheduler.script = wdir/jobscript_fname

            self.scheduler.user_commands = "gdp -p {} compute {}\n".format(
                (wdir/f"worker-{uid}.yaml").name, dataset_path
            )

            # - TODO: check whether params for scheduler is changed
            self.scheduler.write()
            if self._submit:
                self._print(f"{wdir.name} JOBID: {self.scheduler.submit()}")
            else:
                self._print(f"{wdir.name} waits to submit.")

        # - save this batch job to the database
        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            _ = database.insert(
                dict(
                    uid = uid,
                    #md5 = identifier,
                    gdir=job_name, 
                    #group_number=ig, 
                    wdir_names=[str(wdir)], 
                    queued=True
                )
            )

        return

    def inspect(self, resubmit=False, *args, **kwargs):
        """"""
        self._initialise(*args, **kwargs)
        self._debug(f"~~~{self.__class__.__name__}+inspect")

        running_jobs = self._get_running_jobs() # Always return one job
        self._debug(f"running_jobs: {running_jobs}")

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in running_jobs:
                doc_data = database.get(Query().gdir == job_name)
                uid = doc_data["uid"]
                wdir_name = pathlib.Path(doc_data["wdir_names"][0]).name

                self.scheduler.job_name = job_name
                self.scheduler.script = self.directory/f"run-{uid}.script"

                if self.scheduler.is_finished():
                    # -- check if the job finished properly
                    # assert wdir_name = self.wdir_name
                    self.driver.directory = self.directory/wdir_name
                    if self.driver.read_convergence():
                        database.update({"finished": True}, doc_ids=[doc_data.doc_id])
                    else:
                        if resubmit:
                            jobid = self.scheduler.submit()
                            self._print(f"{job_name} is re-submitted with JOBID {jobid}.")
                else:
                    self._print(f"{job_name} is running...")

        return

    def retrieve(
        self, include_retrieved: bool=False, use_archive: bool=False,
        *args, **kwargs
    ):
        """Retrieve training results.

        Args:
            use_archive: Whether archive finished computation folders.

        """
        self.inspect(*args, **kwargs)
        self._debug(f"~~~{self.__class__.__name__}+retrieve")

        unretrieved_wdirs_ = []
        if not include_retrieved:
            unretrieved_jobs = self._get_unretrieved_jobs()
        else:
            unretrieved_jobs = self._get_finished_jobs()

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                unretrieved_wdirs_.extend(
                    [pathlib.Path(w).resolve() for w in doc_data["wdir_names"]]
                )
            self._debug(f"{unretrieved_wdirs_ = }")
            if self._retrieve_mode == "all":
                unretrieved_wdirs = [p for p in unretrieved_wdirs_]
            elif self._retrieve_mode == "single":
                unretrieved_wdirs = [p for p in unretrieved_wdirs_ if p.name == self.wdir_name]
            else:
                ... # The retreive mode shoul be checked before.

        # NOTE: Computation folders should have the same name convention here!
        existed_wdirs = list([x.resolve() for x in self.directory.glob(f"{self.COMP_PREFIX}*")])
        unretrieved_wdirs = [x for x in unretrieved_wdirs if x in existed_wdirs]

        self._debug(f"{existed_wdirs = }")
        self._debug(f"{unretrieved_wdirs = }")
        self._debug(f"{self.directory = }")
        self._debug(f"{self.wdir_name = }")

        results = []
        if unretrieved_wdirs:
            unretrieved_wdirs = [pathlib.Path(x) for x in unretrieved_wdirs]
            # - Find existed computation folders!!
            is_archived = False

            archive_path = (self.directory/"cand.tgz").absolute()
            if not archive_path.exists():
                unretrieved_and_unarchived_wdirs = unretrieved_wdirs
                results = self._read_results(unretrieved_wdirs, )
            else:
                # NOTE: Find previously archived wdirs
                archived_wdirs = []
                with tarfile.open(archive_path, "r:gz") as tar:
                    for tarinfo in tar:
                        if tarinfo.isdir() and tarinfo.name.startswith(self.COMP_PREFIX):
                            archived_wdirs.append(self.directory / tarinfo.name)
                        else:
                            ...
                archived_wdirs = sorted(archived_wdirs, key=lambda x: int(x.name[len(self.COMP_PREFIX) :]))
                self._debug(f"{archived_wdirs = }")
                # TODO: Let driver determines when the computation folder is archived or
                #       not...
                # TODO: Deal with a situation where archived and unarchived ones are mixed?
                unretrieved_and_unarchived_wdirs = [x for x in unretrieved_wdirs if x not in archived_wdirs]
                if len(unretrieved_and_unarchived_wdirs) > 0:
                    is_archived = False
                else:
                    is_archived = True
                if is_archived:
                    results = self._read_results(unretrieved_wdirs, archive_path)
                else:
                    results = self._read_results(unretrieved_wdirs)
            # - archive results if it has not been done
            if use_archive and not is_archived:
                self._print("archive computation folders...")
                if not archive_path.exists():
                    #with tarfile.open(archive_path, "w:gz") as tar:
                    #    for w in unretrieved_wdirs:
                    #        tar.add(w, arcname=w.name)
                    archive_data = io.BytesIO()
                    # -- append
                    with tarfile.open(fileobj=archive_data, mode="w") as tar:
                        for w in unretrieved_wdirs:
                            self._debug(f"add {w.name} to archive.")
                            tar.add(w, arcname=w.name)
                    archive_data.seek(0)
                else:
                    # -- load
                    archive_data = io.BytesIO()
                    with gzip.open(archive_path, "rb") as gzf:
                        archive_data.write(gzf.read())
                    archive_data.seek(0)
                    # -- append
                    with tarfile.open(fileobj=archive_data, mode="a") as tar:
                        for w in unretrieved_and_unarchived_wdirs:
                            self._debug(f"add {w.name} to archive.")
                            tar.add(w, arcname=w.name)
                    archive_data.seek(0)
                # -- save archive
                with gzip.open(archive_path, mode="wb", compresslevel=6) as gzf:
                    gzf.write(archive_data.read())
                for w in unretrieved_and_unarchived_wdirs:
                    shutil.rmtree(w)
            else:
                ... # Nothing to archive
        else:
            ... # Nothing to retrieve

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                database.update({"retrieved": True}, doc_ids=[doc_data.doc_id])

        return results
    
    def _read_results(
            self, unretrieved_wdirs: List[pathlib.Path], archive_path: pathlib.Path=None,
            *args, **kwargs
        ):
        """"""
        # NOTE: SingleWorker always have one unretrieved directory...
        #       However, the retrieve mode can be set to all to get all computation
        #       results in the current main working directory `self.directory`.
        # TODO: Add joblib!!!
        results = []
        for p in unretrieved_wdirs: 
            self.driver.directory = p
            results.append(self.driver.read_trajectory(archive_path=archive_path))
        
        # TODO: dump results shape? assume the results are 2D
        #if self._retrieve_mode == "all":
        #    dim1 = len(results)
        #    dim2 = max([len(x) for x in results])
        #    size = (dim1, dim2)
        #    markers = []
        #    for i in range(dim1):
        #        maxj = len(results[i])
        #        markers.extend([(i, j) for j in range(maxj)])

        #    #shape_dir = self.directory / "_shape"
        #    shape_dir = self.directory.parent / "_shape"
        #    shape_dir.mkdir(parents=True, exist_ok=True)
        #    np.savetxt(shape_dir / "shape.dat", np.array(size, dtype=np.int32), fmt="%8d")
        #    np.savetxt(shape_dir / "markers.dat", np.array(markers, dtype=np.int32), fmt="%8d")

        return results
    
    def as_dict(self) -> dict:
        """"""
        params = super().as_dict()
        params["use_single"] = True

        return params
    

if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import dataclasses
from typing import List, Callable

import numpy as np

import jax
import jax.numpy as jnp

from ase.calculators.calculator import Calculator, all_changes


@jax.jit
def harmonic(positions, cvfunc: Callable, k: float, s: float):
    """Harmonic bias potential."""
    colvars = cvfunc(positions) # NOTE: Must be 1D CV.
    bias = k*jnp.square(colvars - s)

    return jnp.sum(bias)

@jax.jit
def upper_harmonic(positions, cvfunc: Callable, k: float, s: float):
    """Harmonic bias potential."""
    colvars = cvfunc(positions) # NOTE: Must be 1D CV.
    colvars = jnp.where(colvars > s, colvars, s)

    bias = k*jnp.square(colvars - s)

    return jnp.sum(bias)

@jax.jit
def lower_harmonic(positions, cvfunc: Callable, k: float, s: float):
    """Harmonic bias potential."""
    colvars = cvfunc(positions) # NOTE: Must be 1D CV.
    colvars = jnp.where(colvars < s, colvars, s)

    bias = k*jnp.square(colvars - s)

    return jnp.sum(bias)



@dataclasses.dataclass
class HarmonicSetting:

    #: Spring constant, unit of eV.
    k: float = 100.

    #: Function origin, unit of colvar.
    s: float = 0.

    #: Whether construct a upper wall.
    upper: bool = False

    #: Whether construct a lower wall.
    lower: bool = False


class HarmonicBias(Calculator):

    implemented_properties = ["energy", "free_energy", "forces"]

    default_parameters = dict()

    """A harmonic bias.

    E_bias = k*(s_i-s_0)**2

    """

    def __init__(self, colvar: dict=None, restart=None, label=None, atoms=None, directory=".", **kwargs):
        """"""
        super().__init__(restart=restart, label=label, atoms=atoms, directory=directory, **kwargs)

        # - check bias params
        self._setting = HarmonicSetting(**self.parameters)
        assert not (self._setting.upper and self._setting.lower)

        bias_func = harmonic
        if self._setting.upper:
            bias_func = upper_harmonic
        if self._setting.lower:
            bias_func = lower_harmonic
        self.compute_harmonic = jax.value_and_grad(bias_func, argnums=0)

        # - check colvar
        self.colvar = initiate_colvar(colvar)

        return
    
    def calculate(self, atoms=None, properties=["energy"], system_changes=["positions"]):
        """"""
        super().calculate(atoms, properties, system_changes)

        positions = jnp.array(atoms.get_positions())
        ret = self.compute_harmonic(
            positions, self.colvar,
            k=self._setting.k, s=self._setting.s, 
        )
        self.results["energy"] = np.asarray(ret[0])
        self.results["forces"] = -np.array(ret[1]) # copy forces

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
from pathlib import Path
import pickle

from typing import List

import numpy as np

from joblib import Parallel, delayed

import networkx as nx

from ase import Atoms
from ase.io import read, write

from gdpx.utils.command import parse_input_file

from gdpx.graph.creator import StruGraphCreator
from gdpx.graph.sites import SiteGraphCreator
from gdpx.graph.utils import unique_chem_envs, compare_chem_envs
from gdpx.graph.para import paragroup_unique_chem_envs
from gdpx.graph.utils import plot_graph, unpack_node_name

def para_cmp_graphs():
    # --- compare chem envs - parallel
    """
    pair_indices = []
    for i in range(nframes):
        for j in range(i+1,nframes):
            pair_indices.append((i,j))
    print("number of comparasions: ", len(pair_indices))
    cmp_results = Parallel(n_jobs=n_jobs)(delayed(compare_chem_envs)(chem_groups[i], chem_groups[j]) for i, j in pair_indices)

    # merge results
    last_idx = nframes - 1
    found_last = False
    end_idx = np.cumsum(np.arange(nframes-1,0,-1))
    start_idx = [0] + list(end_idx)[:-1]
    unique_groups = []
    for i, (s, e) in enumerate(zip(start_idx, end_idx)):
        for x in unique_groups:
            if i in x:
                break
        else:
            cur_res = cmp_results[s:e]
            cur_pairs = pair_indices[s:e]
            ug = [i]
            ug += [cur_pairs[j][1] for j, x in enumerate(cur_res) if x]
            if len(ug) == 0:
                ug = [i]
            if last_idx in ug:
                found_last = True
            unique_groups.append(ug)
            print(ug)
    # check last
    if not found_last:
        ug.append([last_idx])
    # reformat
    unique_groups = [[[m, frames[m]]for m in x] for x in unique_groups]

    et = time.time()
    print("*time* cmp chem envs: ", et - st)
    """

    return


def add_two():
    xyz_path = "/users/40247882/scratch2/oxides/graph/MyCode/s44s-O1.xyz"
    atoms = read(xyz_path, "16")
    print(atoms)

    site_creator = SiteGraphCreator(
        pbc_grid = [2, 2, 0],
        covalent_ratio = 1.1,
        skin = 0.25,
        adsorbate_elements = ["O"],
        coordination_numbers = [3],
        site_radius=3
    )
    sites = site_creator.convert_atoms(atoms, True)

    print("unique groups: ", len(sites))
    print([len(s) for s in sites])

    print("\n\n===== check site =====")
    ads = Atoms("O", positions=[[0, 0, 0]])
    frames = []
    for ig, sg in enumerate(sites):
        for s in sg: 
            print(s)
            #s.is_occupied(["O"])

            #print(s.known)
            new_atoms = s.adsorb(ads, [64], height=1.3)
            if not isinstance(new_atoms, Atoms):
                print("!!! site may be already occupied!!!", new_atoms)
            else:
                frames.append(new_atoms)
    write("s44s-O2-2.xyz", frames)

    return

def check_unique():
    """"""
    xyz_path = "/mnt/scratch2/users/40247882/oxides/graph/s44s-O2.xyz"
    frames = read(xyz_path, ":")

    stru_creator = StruGraphCreator(
        pbc_grid = [2, 2, 0],
        graph_radius = 1,
        covalent_ratio = 1.1,
        skin = 0.25,
        adsorbate_elements = ["O"]
    )

    def cmp_two(a, b, frames):
        print(f"compare index {a} and {b}")
        frames = [frames[a], frames[b]]

        # check f1 and f2
        atoms1 = frames[0]
        _ = stru_creator.generate_graph(atoms1)
        chem_envs_1 = stru_creator.extract_chem_envs()

        plot_graph(chem_envs_1[0], fig_name=f"ads-1.png")

        atoms2 = frames[1]
        _ = stru_creator.generate_graph(atoms2)
        chem_envs_2 = stru_creator.extract_chem_envs()

        plot_graph(chem_envs_2[0], fig_name=f"ads-2.png")

        print("comparasion: ", compare_chem_envs(chem_envs_1, chem_envs_2))
    
    #cmp_two(8, 9, frames)
    #exit()

    chem_groups = []
    for i, atoms in enumerate(frames):
        print(f"-x-x-x- check frame {i} -x-x-x-")
        _ = stru_creator.generate_graph(atoms)
        chem_envs = stru_creator.extract_chem_envs()
        chem_groups.append(chem_envs)

        print("number of adsorbate graphs: ", len(chem_envs))
    
    unique_envs, unique_groups = unique_chem_envs(
        chem_groups, list(enumerate(frames))
    )
    print("number of unique groups: ", len(unique_groups))

    unique_frames = []
    for ug in unique_groups:
        print("{} {} {}".format(ug[0][0], ug[0][1], len(ug)-1))
        for duplicate in ug[1:]:
            print(duplicate[0])

    write("unique.xyz", unique_frames)

    return

def add_adsorbate_single(site_creator, frames):
    created_frames = []
    for i, atoms in enumerate(frames):
        print(f"====== create sites {i} =====")
        sites = site_creator.convert_atoms(atoms, check_unique=True)

        print("unique sites: ", len(sites))
        print([len(s) for s in sites])
        print("nsites: ", sum([len(s) for s in sites]))

        print("\n\n===== check site =====")
        ads = Atoms("O", positions=[[0, 0, 0]]) # TODO: make this an input structure

        for ig, sg in enumerate(sites):
            # NOTE: only choose unique site
            print("number of uniques: ", len(sg))
            for s in sg[:1]: 
                print(s)

                new_atoms = s.adsorb(ads, site_creator.ads_indices, height=1.3)
                if not isinstance(new_atoms, Atoms):
                    print(s, "!!! site may be already occupied!!!", new_atoms)
                else:
                    new_atoms.info["cycle"] = s.cycle
                    new_atoms.arrays["order"] = np.array(range(len(new_atoms)))
                    created_frames.append(new_atoms)

    write("created-O3.xyz", created_frames)

    return

def create_structure_graphs(input_dict, idx, atoms, selected_indices=None):
    """"""
    stru_creator = StruGraphCreator(
        **input_dict
    )
    _ = stru_creator.generate_graph(atoms, selected_indices)
    #if selected_indices is None:
    #    chem_envs = stru_creator.extract_chem_envs(atoms)
    #else:
    #    chem_envs = stru_creator.extract_selected_chem_envs(selected_indices=selected_indices)
    chem_envs = stru_creator.extract_chem_envs(atoms)

    print(f"check frame {idx}")
    print("number of adsorbate graphs: ", len(chem_envs))

    return chem_envs


def add_adsorbate(input_dict, idx, atoms, ads, distance_to_site=1.5, check_unique=False, pfunc=print):
    """Insert adsorbate into the graph."""
    #print(f"====== create sites {i} =====")
    pfunc("check_unique in sites: ", check_unique)
    print(input_dict)

    site_creator = SiteGraphCreator(**input_dict)
    sites = site_creator.convert_atoms(atoms, check_unique=check_unique) # TODO: custom?

    noccupied = 0

    created_frames = []
    for ig, sg in enumerate(sites):
        # NOTE: only choose unique site
        #print("number of uniques: ", len(sg))
        if check_unique:
            sg = [sg[0]]
        for s in sg: 
            new_atoms = s.adsorb(ads, site_creator.graph_creator.ads_indices, distance_to_site=distance_to_site)
            if not isinstance(new_atoms, Atoms):
                noccupied += 1
                #print(s, "!!! site may be already occupied!!!", new_atoms)
            else:
                new_atoms.info["cycle"] = s.site_indices
                new_atoms.arrays["order"] = np.array(range(len(new_atoms)))
                created_frames.append(new_atoms)

    content = f"===== frame {idx} =====\n"
    content += "unique sites: %d\n" %len(sites)
    content += str([len(s) for s in sites]) + "\n"
    content += "nsites: %d\n" %sum([len(s) for s in sites])
    content += "noccupied: %d\n" %noccupied
    content += "\n\n"
    pfunc(content)
    
    return created_frames

def graph_main(n_jobs, graph_input_file, stru_path, indices, choice):
    """ sift unique adsorbate structures
        frames must have same chemical formula
    """
    print(f"*** number of processors {n_jobs} ***")

    input_dict = parse_input_file(graph_input_file)
    print(input_dict)

    frames = read(stru_path, indices)
    nframes = len(frames)

    print(f"===== number of frames {nframes}=====")

    if choice == "diff": # analyse chemical environment and output unique structures
        stru_creator = StruGraphCreator(
            **input_dict
        )

        # calculate chem envs
        st = time.time()

        p = Path("chemenvs.pkl")
        if not p.exists():
            chem_groups = Parallel(n_jobs=n_jobs)(delayed(create_structure_graphs)(input_dict, idx,a) for idx, a in enumerate(frames))
            with open("chemenvs.pkl", "wb") as fopen:
                pickle.dump(chem_groups, fopen)
        else:
            with open("chemenvs.pkl", "rb") as fopen:
                chem_groups = pickle.load(fopen)

        et = time.time()
        print("*time* calc chem envs: ", et - st)
    
        # compare chem envs - serial
        #unique_envs, unique_groups = unique_chem_envs(chem_groups, list(enumerate(frames)))
        #unique_envs, unique_groups = para_unique_chem_envs(
        #    chem_groups, list(enumerate(frames)), n_jobs=n_jobs
        #)
        #unique_envs, unique_groups 
        unique_envs, unique_groups = paragroup_unique_chem_envs(chem_groups, list(enumerate(frames)), n_jobs=n_jobs)
        #unique_envs, unique_groups = newnew_para_unique_chem_envs(chem_groups, list(enumerate(frames)), n_jobs=n_jobs)
        #unique_envs, unique_groups = block_para_unique_chem_envs(chem_groups, list(enumerate(frames)), n_jobs=n_jobs)
        #unique_envs, unique_groups = pool_para_unique_chem_envs(chem_groups, list(enumerate(frames)), n_jobs=n_jobs)
        #unique_envs, unique_groups = new_pool_para_unique_chem_envs(chem_groups, list(enumerate(frames)), n_jobs=n_jobs)

        et = time.time()
        print("*time* cmp chem envs: ", et - st)

        #new_inputs = [[chem_groups, list(enumerate(frames))]]*8
        #xxx = Parallel(n_jobs=n_jobs)(delayed(unique_chem_envs)(a1, b1) for a1, b1 in new_inputs)
        #print("*time* cmp chem envs: ", et - st)

        print("number of unique groups: ", len(unique_groups))

        unique_data = []
        for i, x in enumerate(unique_groups):
            data = ["ug"+str(i)]
            data.extend([a[0] for a in x])
            unique_data.append(data)
        content = "# unique, indices\n"
        content += f"# ncandidates {nframes}\n"
        for d in unique_data:
            content += ("{:<8s}  "+"{:<8d}  "*(len(d)-1)+"\n").format(*d)

        with open(Path.cwd() / "unique-g.txt", "w") as fopen:
            fopen.write(content)

        all_unique = []
        print("\n# uniqueid nduplicates")
        for iu, ug in enumerate(unique_groups):
            unique_frames = []
            for x in ug:
                x[1].info["confid"] = x[0]
                unique_frames.append(x[1])
            unique_frames = sorted(unique_frames, key=lambda a: a.get_potential_energy(), reverse=False)
            #write(f"unique-g{iu}.xyz", unique_frames)
            all_unique.append(unique_frames[0])
            #print("{} {} {}".format(ug[0][0], ug[0][1], len(ug)-1))
            #print("{} {}".format(ug[0][0], len(ug)-1))
            #for duplicate in ug[1:]:
            #    print(duplicate[0])
        write(Path.cwd() / "ug-candidates.xyz", all_unique)

    elif choice == "add":
        site_creator = SiteGraphCreator(
            **input_dict
        )
        print("input surface mask: ", site_creator.surface_mask)

        # TODO: how to use clean? 
        print("input clean: ", site_creator.clean)

        # joblib version
        st = time.time()

        ads = Atoms("O", positions=[[0, 0, 0]]) # TODO: make this an input structure

        ads_frames = Parallel(n_jobs=n_jobs)(delayed(add_adsorbate)(input_dict, idx, a, ads) for idx, a in enumerate(frames))
        #print(ads_frames)

        created_frames = []
        for af in ads_frames:
            created_frames.extend(af)
        write("created-O3.xyz", created_frames)

        et = time.time()
        print("time: ", et - st)

    else:
        pass

    return

if __name__ == "__main__":
    add_two()
    #check_unique()
    pass


""" 

"""

import numpy as np

from ase.md.md import MolecularDynamics
from ase.md.velocitydistribution import Stationary
from ase import units

def force_temperature(atoms, temperature, unit="K"):
    """ force (nucl.) temperature to have a precise value

    Parameters:
    atoms: ase.Atoms
        the structure
    temperature: float
        nuclear temperature to set
    unit: str
        'K' or 'eV' as unit for the temperature
    """

    eps_temp = 1e-12

    if unit == "K":
        E_temp = temperature * units.kB
    elif unit == "eV":
        E_temp = temperature
    else:
        raise UnitError("'{}' is not supported, use 'K' or 'eV'.".format(unit))

    # check DOF
    ndof = 3*len(atoms)
    for constraint in atoms._constraints:
        ndof -= constraint.get_removed_dof(atoms)

    # calculate kinetic energy and get the scale
    if temperature > eps_temp:
        E_kin0 = atoms.get_kinetic_energy() / (0.5 * ndof)
        gamma = E_temp / E_kin0
    else:
        gamma = 0.0
    
    atoms.set_momenta(atoms.get_momenta() * np.sqrt(gamma))

    return




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
import re

import yaml

from ..core.variable import Variable, DummyVariable
from ..core.operation import Operation
from ..core.register import registers

from ..potential.manager import AbstractPotentialManager
from ..potential.trainer import AbstractTrainer
from ..worker.train import TrainerBasedWorker
from ..scheduler.interface import SchedulerVariable
from ..scheduler.scheduler import AbstractScheduler


@registers.variable.register
class TrainerVariable(Variable):

    def __init__(self, directory="./", **kwargs):
        """"""
        # print("trainer keys: ", kwargs.keys())
        name = kwargs.get("name", None)
        trainer = registers.create("trainer", name, convert_name=True, **kwargs)

        super().__init__(initial_value=trainer, directory=directory)

        return


@registers.operation.register
class train(Operation):

    #: Whether to actively update some attrs.
    _active: bool = False

    def __init__(
        self,
        dataset,
        trainer,
        potter,
        scheduler=DummyVariable(),
        size: int = 1,
        init_models=None,
        active: bool = False,
        share_dataset: bool = False,
        auto_submit: bool = True,
        directory="./",
        *args,
        **kwargs,
    ) -> None:
        """"""
        input_nodes = [dataset, trainer, scheduler, potter]
        super().__init__(input_nodes=input_nodes, directory=directory)

        assert (
            trainer.value.name == potter.value.name
        ), "Trainer and potter have inconsistent name."
        assert (
            trainer.value.type_list == potter.value.as_dict()["params"]["type_list"]
        ), "Trainer and potter have inconsistent type_list."

        self.size = size  # number of models
        if init_models is not None:
            self.init_models = [str(pathlib.Path(p).absolute()) for p in init_models]
        else:
            self.init_models = [None] * self.size
        assert (
            len(self.init_models) == self.size
        ), f"The number of init models {self.init_models} is inconsistent with size {self.size}."

        self._active = active

        self._share_dataset = share_dataset
        self._auto_submit = auto_submit

        return

    def forward(
        self,
        dataset,
        trainer: AbstractTrainer,
        scheduler: AbstractScheduler,
        potter: AbstractPotentialManager,
    ):
        """"""
        super().forward()

        init_models = self.init_models
        if self._active:
            curr_iter = int(self.directory.parent.name.split(".")[-1])
            if curr_iter > 0:
                self._print(">>> Update init_models...")
                prev_wdir = (
                    self.directory.parent.parent
                    / f"iter.{str(curr_iter-1).zfill(4)}"
                    / self.directory.name
                )
                prev_mdirs = []  # model dirs
                for p in prev_wdir.iterdir():
                    if p.is_dir() and re.match("m[0-9]+", p.name):
                        prev_mdirs.append(p)
                # TODO: replace `m` with a constant
                init_models = []
                prev_mdirs = sorted(prev_mdirs, key=lambda p: int(p.name[1:]))
                for p in prev_mdirs:
                    trainer.directory = p
                    if hasattr(trainer, "get_checkpoint"):
                        init_models.append(trainer.get_checkpoint())
                    else:
                        init_models.append((p / trainer.frozen_name).resolve())
                for p in init_models:
                    self._print(f"  {str(p)}")
                assert init_models, "No previous models found."

        # -
        if scheduler is None:
            scheduler = SchedulerVariable().value

        # - update dir
        worker = TrainerBasedWorker(
            trainer,
            scheduler,
            share_dataset=self._share_dataset,
            auto_submit=self._auto_submit,
            directory=self.directory,
        )

        # - run
        manager = None

        _ = worker.run(dataset, size=self.size, init_models=init_models)
        _ = worker.inspect(resubmit=True)
        if worker.get_number_of_running_jobs() == 0:
            models = worker.retrieve(include_retrieved=True)
            self._print("Frozen Models: ")
            for m in models:
                self._print(f"  {str(m) =}")
            potter_params = potter.as_dict()
            potter_params["params"]["model"] = models
            potter.register_calculator(potter_params["params"])
            manager = potter
        else:
            self._print("TrainWorker has not finished.")

        if manager is not None:
            self.status = "finished"
        
        # - some imported packages change `logging.basicConfig`
        #   and accidently add a StreamHandler to logging.root
        #   so remove it...
        import logging
        for h in logging.root.handlers:
            if isinstance(h, logging.StreamHandler) and not isinstance(
                h, logging.FileHandler
            ):
                logging.root.removeHandler(h)

        return manager


@registers.operation.register
class save_potter(Operation):

    def __init__(self, potter, dst_path=None, directory="./") -> None:
        """"""
        input_nodes = [potter]
        super().__init__(input_nodes, directory)

        if dst_path is not None:
            self.dst_path = pathlib.Path(dst_path).absolute()
            suffix = self.dst_path.suffix
            assert suffix == ".yaml", "dst_path should be either a yaml or a json file."
        else:
            self.dst_path = self._output_path

        return

    def forward(self, potter):
        """"""
        super().forward()

        self._output_path = self.directory / "potter.yaml"
        with open(self._output_path, "w") as fopen:
            yaml.safe_dump(potter.as_dict(), fopen, indent=2)

        if self.dst_path.exists():
            self._print("remove previous potter...")
            self.dst_path.unlink()
        self.dst_path.symlink_to(self._output_path)

        self.status = "finished"

        return potter


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List, Union
from itertools import groupby
from operator import itemgetter

import numpy as np

from ase.constraints import constrained_indices, FixAtoms


def convert_indices(indices: Union[str,List[int]], index_convention="lmp"):
    """ parse indices for reading xyz by ase, get start for counting
        constrained indices followed by lammps convention
        "2:4 3:8"
        convert [1,2,3,6,7,8] to "1:3 6:8"
        lammps convention starts from 1 and includes end
        ---
        input can be either py or lmp
        output for indices is in py since it can be used to access atoms
        output for text is in lmp since it can be used in lammps or sth
    """
    ret = []
    if isinstance(indices, str):
        # string to List[int]
        for x in indices.strip().split():
            cur_range = list(map(int, x.split(":")))
            if len(cur_range) == 1:
                start, end = cur_range[0], cur_range[0]
            else:
                start, end = cur_range
            if index_convention == "lmp":
                ret.extend([i-1 for i in list(range(start,end+1))])
            elif index_convention == "py":
                ret.extend(list(range(start,end)))
            else:
                pass
    elif isinstance(indices, list):
        # List[int] to string
        indices = sorted(indices)
        if index_convention == "lmp":
            pass
        elif index_convention == "py":
            indices = [i+1 for i in indices]
        ret = []
        #ranges = []
        for k, g in groupby(enumerate(indices),lambda x:x[0]-x[1]):
            group = (map(itemgetter(1),g))
            group = list(map(int,group))
            #ranges.append((group[0],group[-1]))
            if group[0] == group[-1]:
                ret.append(str(group[0]))
            else:
                ret.append("{}:{}".format(group[0],group[-1]))
        ret = " ".join(ret)
    else:
        pass

    return ret

def parse_constraint_info(atoms, cons_text, ignore_ase_constraints=True, ret_text=True) -> List[int]:
    """ constraint info can be any forms below, 
        and transformed into indices that start from 1
        "2:5 8" means 2,3,4,5,8 (default uses lmp convention)
        "py 0:5 9" means 1,2,3,4,5,10
        "lmp 1:4 8" means 1,2,3,4,8
        "lowest 10" means 10 atoms with smallest z-positions
        "zpos 4.5" means all atoms with zpositions smaller than 4.5

        return lammps format text or python format list
    """
    # - set some init values
    aindices = list(range(len(atoms)))
    mobile_indices = aindices.copy()
    frozen_indices = []
    #print("constraint: ", cons_text)

    # TODO: check if atoms have constraint
    # NOTE: only need to determine which atoms are frozen then others are mobile
    cons_indices = constrained_indices(atoms, only_include=FixAtoms) # array
    if (not ignore_ase_constraints) and cons_indices.size > 0:
        # convert to lammps convention
        frozen_indices = cons_indices.copy().tolist()
    else:
        # TODO: if use region indicator
        if cons_text is None:
            pass
        else:
            # - parse constraint text type
            cons_data = cons_text.split()
            if cons_data[0] not in ["py", "lmp", "lowest", "zpos"]:
                cons_type, cons_info = "lmp", " ".join(cons_data)
            else:
                cons_type, cons_info = cons_data[0], " ".join(cons_data[1:])
            #print("cons_info: ", cons_type, cons_info)

            # NOTE: text may have different notations
            #       but indices should all be in python convention
            if cons_type == "py":
                frozen_indices = convert_indices(cons_info, index_convention="py")
            elif cons_type == "lmp":
                frozen_indices = convert_indices(cons_info, index_convention="lmp")
            elif cons_type == "lowest":
                if atoms.pbc[2]:
                    z_coordinates = atoms.get_scaled_positions()[:, 2]
                    wrapped_coordinates = []
                    for z in z_coordinates: # wrap to [-0.1, 0.9)
                        z_ = np.modf(z)[0]
                        if z_ < -0.1:
                            z_ += 1.0
                        elif z >= 0.9:
                            z_ -= 1.0
                        wrapped_coordinates.append(z_)
                    frozen_indices = sorted(aindices, key=lambda x: wrapped_coordinates[x])[:int(cons_info)]
                else:
                    frozen_indices = sorted(aindices, key=lambda x:atoms.positions[x][2])[:int(cons_info)]
            elif cons_type == "zpos":
                frozen_indices = [i for i in aindices if atoms.positions[i][2] <= float(cons_info)]
            else:
                pass
    mobile_indices = [i for i in aindices if i not in frozen_indices]
    #print(mobile_indices)
    #print(frozen_indices)

    if ret_text:
        frozen_text = convert_indices(frozen_indices, index_convention="py")
        mobile_text = convert_indices(mobile_indices, index_convention="py")

        return mobile_text, frozen_text
    else:
        return mobile_indices, frozen_indices

def set_constraint(atoms, cons_text, ignore_attached_constraints=True):
    """Set constraint based on constraint text."""
    mobile_indices, frozen_indices = parse_constraint_info(
        atoms, cons_text, ignore_ase_constraints=ignore_attached_constraints, ret_text=False
    )
    if frozen_indices:
        if ignore_attached_constraints:
            atoms._del_constraints()
        else:
            ...
        atoms.set_constraint(FixAtoms(indices=frozen_indices))
    else:
        ...

    return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import NoReturn, List

import numpy as np
import jax.numpy as jnp

from ase import Atoms
from ase.calculators.calculator import Calculator


class AbstractBias(Calculator):

    implemented_properties = ["energy", "free_energy", "forces"]

    default_parameters = dict()

    def __init__(self, colvars: List[dict]=None, restart=None, label=None, atoms=None, directory=".", **kwargs):
        """"""
        super().__init__(restart=restart, label=label, atoms=atoms, directory=directory, **kwargs)

        # - check colvar
        colvars_ = []
        if isinstance(colvars, list):
            for colvar in colvars:
                colvars_.append(initiate_colvar(colvar))
        elif isinstance(colvars, dict):
            colvars_.append(initiate_colvar(colvars))
        else:
            ...
        self.colvars = colvars_

        # - NOTE: set bias function and parameters in subclass!
        ...

        return

    def calculate(self, atoms=None, properties=["energy"], system_changes=["positions"]):
        """"""
        super().calculate(atoms, properties, system_changes)

        positions = jnp.array(atoms.get_positions())
        bias_energy, bias_forces = 0., np.zeros(positions.shape)
        for colvar, bias_params in zip(self.colvars, self.bias_params):
            ret = self._compute_bias(
                positions, colvar, **bias_params
            )
            bias_energy += np.asarray(ret[0])
            bias_forces += -np.array(ret[1])
        self.results["energy"] = bias_energy
        self.results["forces"] = bias_forces

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import logging
import pathlib
from typing import Optional, Union, List

from ase.io import write

from ..worker.drive import DriverBasedWorker
from ..worker.interface import ComputerVariable, ReactorVariable
from ..reactor.reactor import AbstractReactor
from ..utils.command import parse_input_file

DEFAULT_MAIN_DIRNAME = "MyWorker"


def convert_config_to_potter(config):
    """Convert a configuration file or a dict to a potter/reactor.

    This function is only called in the `main.py`.

    """
    if isinstance(config, dict):
        params = config
    else:  # assume it is json or yaml
        params = parse_input_file(input_fpath=config)

    ptype = params.pop("type", "computer")

    # NOTE: compatibility
    potter_params = params.pop("potter", None)
    potential_params = params.pop("potential", None)
    if potter_params is None:
        if potential_params is not None:
            params["potter"] = potential_params
        else:
            raise RuntimeError("Fail to find any potter (potential) definition.")
    else:
        params["potter"] = potter_params

    if ptype == "computer":
        potter = ComputerVariable(**params).value
    elif ptype == "reactor":

        potter = ReactorVariable(
            potter=params["potter"],
            driver=params.get("driver", None),
            scheduler=params.get("scheduler", {}),
            batchsize=params.get("batchsize", 1),
        ).value[0]
    else:
        ...

    return potter

def run_one_worker(structures, worker, directory, batch, spawn, archive):
    """"""
    worker.directory = directory

    _ = worker.run(structures, batch=batch)
    worker.inspect(resubmit=True)
    if not spawn and worker.get_number_of_running_jobs() == 0:
        res_dir = directory / "results"
        if not res_dir.exists():
            res_dir.mkdir(exist_ok=True)

            ret = worker.retrieve(include_retrieved=True, use_archive=archive)
            if not isinstance(worker.driver, AbstractReactor):
                end_frames = [traj[-1] for traj in ret]
                write(res_dir / "end_frames.xyz", end_frames)
            else:
                ...
        else:
            print(f"{directory.name} has already been retrieved.")

    return

def run_worker(
    structure: List[str],
    workers: List[DriverBasedWorker],
    *,
    batch: Optional[int] = None,
    spawn: bool = False,
    archive: bool = False,
    directory: Union[str, pathlib.Path] = pathlib.Path.cwd() / DEFAULT_MAIN_DIRNAME,
):
    """"""
    # some imported packages change `logging.basicConfig`
    # and accidently add a StreamHandler to logging.root
    # so remove it...
    for h in logging.root.handlers:
        if isinstance(h, logging.StreamHandler) and not isinstance(
            h, logging.FileHandler
        ):
            logging.root.removeHandler(h)

    # set working directory
    directory = pathlib.Path(directory)
    if not directory.exists():
        directory.mkdir()

    # - read structures
    from gdpx.builder import create_builder

    frames = []
    for i, s in enumerate(structure):
        builder = create_builder(s)
        builder.directory = directory / "init" / f"s{i}"
        frames.extend(builder.run())

    # - find input frames
    num_workers = len(workers)
    if num_workers == 1:
        run_one_worker(frames, workers[0], directory, batch, spawn, archive)
    else:
        for i, w in enumerate(workers):
            run_one_worker(frames, w, directory/f"w{i}", batch, spawn, archive)

    return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import pathlib
import warnings

from typing import List, Mapping
from pathlib import Path

import numpy as np

import ase
from ase import units, Atoms
from ase.io import read, write
from ase.data import covalent_radii

from ase.ga.utilities import get_all_atom_types, closest_distances_generator # get system composition (both substrate and top)
from ase.ga.utilities import CellBounds
from ase.ga.startgenerator import StartGenerator

from gdpx.core.register import registers
from gdpx.builder.builder import StructureBuilder, StructureModifier
from gdpx.builder.species import build_species

""" Generate structures randomly
"""

def compute_molecule_number_from_density(molecular_mass, volume, density) -> int:
    """Compute the number of molecules in the region with a given density.

    Args:
        moleculer_mass: unit in g/mol.
        volume: unit in Ang^3.
        density: unit in g/cm^3.
    
    Returns:
        Number of molecules in the region.

    """
    number = (density/molecular_mass) * volume * units._Nav * 1e-24

    return int(number)


class RandomBuilder(StructureModifier):

    #: Number of attempts to create a random candidate.
    MAX_ATTEMPTS_PER_CANDIDATE: int = 1000

    #: Number of attempts to create a number of candidates.
    #       if 10 structures are to create, run will try 5*10=50 times.
    MAX_TIMES_SIZE: int = 5

    #: Whether use tags to distinguish molecules.
    use_tags: bool = False

    #: Atom numbers of composition to insert.
    composition_atom_numbers: List[int] = None

    #: Composition to insert.
    composition_blocks: Mapping[str,int] = None

    def __init__(
        self, composition: Mapping[str,int], substrates = None,
        region: dict={}, cell=None, covalent_ratio=[0.8, 2.0], 
        max_times_size: int=5, *args, **kwargs
    ):
        super().__init__(substrates=substrates, *args, **kwargs)

        # TODO: substrates should also be a Builder Object
        # TODO: if substrates is a ChemiclFormula?
        if isinstance(substrates, str) or isinstance(substrates, pathlib.Path):
            substrates = str(pathlib.Path(substrates).absolute())
        else:
            ...

        self._state_params = dict(
            composition = composition,
            substrates = substrates,
            region = region,
            cell = cell,
            covalent_ratio = covalent_ratio,
            max_times_size = max_times_size,
            test_too_far = kwargs.get("test_too_far", True), # test_too_far
            test_dist_to_slab = kwargs.get("test_dist_to_slab", True), # test_dist_to_slab
            cell_volume = kwargs.get("cell_volume", None),
            cell_bounds = kwargs.get("cell_bounds", None),
            cell_splits = kwargs.get("cell_splits", None),
            random_seed = self.random_seed
        )

        # - set random seed for generators due to compatibility
        if isinstance(self.random_seed, int):
            np.random.seed(self.random_seed)
        elif isinstance(self.random_seed, dict):
            np.random.set_state(self.random_seed)
        else:
            ...
        
        # - 
        self.MAX_TIMES_SIZE = max_times_size

        # - create region
        region = copy.deepcopy(region)
        shape = region.pop("method", "auto")
        self.region = registers.create("region", shape, convert_name=True, **region)

        # - parse composition
        self.composition = composition
        self._parse_composition()

        self.covalent_min = covalent_ratio[0]
        self.covalent_max = covalent_ratio[1]

        # - add substrate
        self._substrate = None
        if self.substrates is not None:
            self._substrate = self.substrates[0]
        if self._substrate is not None:
            unique_atom_types = get_all_atom_types(self._substrate, self.composition_atom_numbers)
            self.blmin = self._build_tolerance(unique_atom_types)
        else:
            unique_atom_types = set(self.composition_atom_numbers)
            self.blmin = self._build_tolerance(unique_atom_types)

        # - check cell
        self.cell = cell

        # - read from kwargs
        self.test_too_far = kwargs.get("test_too_far", True) # test_too_far
        self.test_dist_to_slab = kwargs.get("test_dist_to_slab", True) # test_dist_to_slab

        self.cell_volume = kwargs.get("cell_volume", None)
        self.cell_bounds = kwargs.get("cell_bounds", None)
        self.cell_splits = kwargs.get("cell_splits", None)
        self.number_of_variable_cell_vectors = 0 # number_of_variable_cell_vectors

        return
    
    def _load_substrates(self, inp_sub) -> List[Atoms]:
        """"""
        substrates = super()._load_substrates(inp_sub)
        if substrates is not None:
            assert len(substrates) == 1, "RandomBuilder only supports one substrate."

        return substrates

    def run(self, substrates: List[Atoms]=None, size: int=1, soft_error: bool=False, *args, **kwargs) -> List[Atoms]:
        """Modify input structures.

        Args:
            substrates: Building blocks.
            numbers: Number of each block.
        
        Returns:
            A list of structures.
        
        """
        super().run(substrates=substrates, *args, **kwargs)

        # - create generator
        generator = self._create_generator(self.substrates)

        # - run over
        frames = []
        for i in range(size*self.MAX_TIMES_SIZE):
            nframes = len(frames)
            if nframes < size:
                atoms = generator.get_new_candidate(maxiter=self.MAX_ATTEMPTS_PER_CANDIDATE)
                if atoms is not None:
                    frames.append(atoms)
                    nframes += 1
            else:
                break
        else:
            if size > 0:
                if soft_error:
                    warnings.warn(f"Failed to create {size} structures, only {nframes} are created.", UserWarning)
                else:
                    raise RuntimeError(f"Failed to create {size} structures, only {nframes} are created.")
            else:
                ...
        
        return frames
    
    def _update_settings(self, substarte: Atoms=None):
        """"""

        raise NotImplementedError()

    def _create_generator(self, substrates: List[Atoms]=None):
        """"""
        if substrates is not None:
            self._update_settings(substrates[0])
        else:
            self._update_settings()

        generator = StartGenerator(
            self._substrate, 
            self.composition_blocks, # blocks
            self.blmin,
            number_of_variable_cell_vectors=self.number_of_variable_cell_vectors,
            box_to_place_in=self.box_to_place_in,
            box_volume=self.cell_volume,
            splits=self.cell_splits,
            cellbounds=self.cell_bounds,
            test_dist_to_slab = self.test_dist_to_slab,
            test_too_far = self.test_too_far,
            rng = np.random
        ) # structure generator

        return generator

    def _parse_composition(self):
        # --- Define the composition of the atoms to optimize ---
        blocks = []
        for k, v in self.composition.items():
            k = build_species(k)
            if isinstance(v, int): # number
                v = v
            else: # string command
                data = v.split()
                if data[0] == "density":
                    v = compute_molecule_number_from_density(
                        np.sum(k.get_masses()), self.region.get_volume(), 
                        density = float(data[1])
                    )
                else:
                    raise RuntimeError(f"Unrecognised composition {k:v}.")
            blocks.append((k,v))
        for k, v in blocks:
            if len(k) > 1:
                self.use_tags = True
                break
        else:
            self.use_tags = False
        self.composition_blocks = blocks

        atom_numbers = [] # atomic number of inserted atoms
        for species, num in self.composition_blocks:
            numbers = []
            for s, n in ase.formula.Formula(species.get_chemical_formula()).count().items():
                numbers.extend([ase.data.atomic_numbers[s]]*n)
            atom_numbers.extend(numbers*num)
        self.composition_atom_numbers = atom_numbers
        
        return

    def _build_tolerance(self, unique_atom_types: List[int]):
        """"""
        blmin = closest_distances_generator(
            atom_numbers=unique_atom_types,
            # be careful with test too far
            ratio_of_covalent_radii = self.covalent_min
        )

        return blmin

    def _print_blmin(self, blmin):
        """"""
        elements = []
        for k in blmin.keys():
            elements.extend(k)
        elements = set(elements)
        #elements = [ase.data.chemical_symbols[e] for e in set(elements)]
        nelements = len(elements)

        index_map = {}
        for i, e in enumerate(elements):
            index_map[e] = i
        distance_map = np.zeros((nelements, nelements))
        for (i, j), dis in blmin.items():
            distance_map[index_map[i], index_map[j]] = dis

        symbols = [ase.data.chemical_symbols[e] for e in elements]

        content =  "Bond Distance Minimum\n"
        content += "  covalent ratio: {}\n".format(self.covalent_min)
        content += "  "+" "*4+("{:>6}  "*nelements).format(*symbols)+"\n"
        for i, s in enumerate(symbols):
            content += "  "+("{:<4}"+"{:>8.4f}"*nelements+"\n").format(s, *list(distance_map[i]))
        content += "  too_far: {}, dist_to_slab: {}\n".format(self.test_too_far, self.test_dist_to_slab)
        content += "  note: default too far tolerance is 2 times\n"

        return content

    def __repr__(self):
        """"""
        content = ""
        content += f"----- {self.__class__.__name__} Parameters -----\n"
        content += f"random_seed: {self.random_seed}\n"

        content += str(self.region)
        if self.blmin is not None:
            content += self._print_blmin(self.blmin)
        else:
            content += "Bond Distance Minimum\n"
            content += "  covalent ratio: {}\n".format(self.covalent_min)

        return content


class BulkBuilder(RandomBuilder):

    name: str = "random_bulk"

    def _update_settings(self, substarte: Atoms = None):
        """"""
        # - ignore substrate
        self._substrate = Atoms("", pbc=True)

        unique_atom_types = set(self.composition_atom_numbers)
        self.blmin = self._build_tolerance(unique_atom_types)

        # - check number_of_variable_cell_vectors
        if self.cell is None:
            self.cell = []
        number_of_variable_cell_vectors = 3 - len(self.cell)
        box_to_place_in = None
        if number_of_variable_cell_vectors > 0:
            box_to_place_in = [[0.,0.,0.], np.zeros((3,3))]
            if len(self.cell) > 0:
                box_to_place_in[1][number_of_variable_cell_vectors:] = self.cell
        self.number_of_variable_cell_vectors = number_of_variable_cell_vectors
        self.box_to_place_in = box_to_place_in

        # --- check volume
        if self.cell_volume is None:
            radii = [covalent_radii[x]*self.covalent_max for x in self.composition_atom_numbers]
            self.cell_volume = np.sum([4/3.*np.pi*r**3 for r in radii])

        # --- cell bounds
        cell_bounds = {}
        angles, lengths = ["phi", "chi", "psi"], ["a", "b", "c"]
        for k in angles:
            cell_bounds[k] = self.cell_bounds.get(k, [15, 165])
        for k in lengths:
            cell_bounds[k] = self.cell_bounds.get(k, [2, 60])
        self.cell_bounds = CellBounds(cell_bounds)

        # --- splits
        if self.cell_splits is not None:
            splits_ = {}
            for r, p in zip(self.cell_splits["repeats"], self.cell_splits["probs"]):
                splits_[tuple(r)] = p
            self.cell_splits = splits_

        return
    
    def as_dict(self) -> dict:
        """"""
        params = copy.deepcopy(self._state_params)
        params["method"] = self.name

        return params


class ClusterBuilder(RandomBuilder):

    name: str = "random_cluster"

    def __init__(
        self, composition: Mapping[str, int], substrates=None, region: dict = {}, 
        cell=None, covalent_ratio=[0.8, 2.0], *args, **kwargs
    ):
        """"""
        super().__init__(
            composition, substrates, region, cell, covalent_ratio, *args, **kwargs
        )

        self.pbc = kwargs.get("pbc", False)

        return

    def _update_settings(self, substarte: Atoms = None):
        """"""
        # - ignore substrate
        if self.cell is None: # None or []
            self.cell = np.array([19.,0.,0.,0.,20.,0.,0.,0.,21.]).reshape(3,3)
        else:
            self.cell = np.reshape(self.cell, (-1,3))
        self._substrate = Atoms(cell = self.cell, pbc=self.pbc)

        if self.region.__class__.__name__ == "AutoRegion":
            from .region import LatticeRegion
            self.region = LatticeRegion(origin=np.zeros(3), cell=self.cell.flatten())

        unique_atom_types = set(self.composition_atom_numbers)
        self.blmin = self._build_tolerance(unique_atom_types)
            
        # - ignore lattice parameters
        self.cell_volume = None
        self.cell_splits = None
        self.cell_bounds = None

        self.box_to_place_in = [self.region._origin, self.region._cell]

        return

    def as_dict(self) -> dict:
        """"""
        params = copy.deepcopy(self._state_params)
        params["method"] = self.name
        params["pbc"] = self.pbc

        return params


class SurfaceBuilder(RandomBuilder):

    name: str = "random_surface"
    
    def __init__(
        self, region: dict, composition: Mapping[str,int], cell=[], 
        covalent_ratio=[1.0, 2.0], *args, **kwargs
    ):
        """"""
        super().__init__(
            region=region, composition=composition, cell=cell, 
            covalent_ratio=covalent_ratio, *args, **kwargs
        )

        # - check region
        assert self.region.__class__.__name__ == "LatticeRegion", f"{self.__class__.__name__} needs a LatticeRegion."

        return
    
    def _update_settings(self, substarte: Atoms = None):
        """"""        
        # - use init substrate
        if substarte is not None:
            self._substrate = substarte

        # define the closest distance two atoms of a given species can be to each other
        unique_atom_types = get_all_atom_types(self._substrate, self.composition_atom_numbers)
        self.blmin = self._build_tolerance(unique_atom_types)

        # - ignore lattice parameters
        self.cell_volume = None
        self.cell_splits = None
        self.cell_bounds = None

        self.box_to_place_in = [self.region._origin, self.region._cell]

        return

    def as_dict(self) -> dict:
        """"""
        params = copy.deepcopy(self._state_params)
        params["method"] = self.name

        return params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import NoReturn

import omegaconf

from ..core.variable import Variable, DummyVariable
from ..core.operation import Operation
from ..core.register import registers

from ..data.array import AtomsNDArray
from ..data.dataset import AbstractDataloader
from ..worker.drive import DriverBasedWorker
from .validator import AbstractValidator

@registers.variable.register
class ValidatorVariable(Variable):

    def __init__(self, directory="./", **kwargs):
        """"""
        # - create a validator
        method = kwargs.get("method", "minima")
        validator = registers.create("validator", method, convert_name=False, **kwargs)

        # - save
        super().__init__(initial_value=validator, directory=directory)

        return
    
@registers.operation.register
class validate(Operation):

    """The operation to validate properties by potentials.

    The reference properties should be stored and accessed through `structures`.

    """

    def __init__(
        self, structures, validator, worker=DummyVariable(), 
        run_params: dict={}, directory="./", *args, **kwargs
    ) -> None:
        """Init a validate operation.

        Args:
            structures: A node that forwards structures.
            validator: A validator.
            worker: A worker to run calculations. TODO: Make this optional.
        
        """
        super().__init__(
            input_nodes=[structures, validator, worker], directory=directory
        )

        self.run_params = run_params

        return
    
    def _preprocess_input_nodes(self, input_nodes):
        """Preprocess valid input nodes.

        Some arguments accept basic python objects such list or dict, which are 
        not necessary to be a Variable or an Operation.

        """
        structures, validator, worker = input_nodes

        if isinstance(validator, dict) or isinstance(validator, omegaconf.dictconfig.DictConfig):
            validator = ValidatorVariable(self.directory/"validator", **validator)
            self._print(validator)

        return structures, validator, worker
    
    def _convert_dataset(self, structures):
        """Validator can accept various formats of input structures. 

        We convert all formats into one.
        
        """
        # NOTE: In an active session, the dataset is dynamic, thus, 
        #       we need load the dataset before run...
        #       Validator accepts dict(reference=[], prediction=[])
        dataset_ = {}
        if hasattr(structures, "items"): # check if the input is a dict-like object
            stru_dict = structures
        else: # assume it is just an AtomsNDArray
            stru_dict = {}
            stru_dict["reference"] = structures

        for k, v in stru_dict.items():
            if isinstance(v, dict):
                ...
            elif isinstance(v, list):
                ...
            elif isinstance(v, AtomsNDArray):
                ...
            elif isinstance(v, AbstractDataloader):
                v = v.load_frames()
            else:
                raise RuntimeError(f"{k} Dataset {type(v)} is not a dict or loader.")
            dataset_[k] = v

        dataset = dataset_

        return dataset
    
    def forward(self, structures, validator: AbstractValidator, workers):
        """Run a validator on input dataset.

        Args:
            structures: Any format that has Atoms objects.

        """
        super().forward()

        # - create a worker
        if workers is not None:
            nworkers = len(workers)
            assert nworkers == 1, f"Validator only accepts one worker but {nworkers} were given."
            worker = workers[0]
            worker.directory = self.directory
        else:
            worker = None

        # - convert dataset
        dataset = self._convert_dataset(structures)

        # - run validation
        validator.directory = self.directory
        validator.run(dataset, worker, **self.run_params)

        self.status = "finished"

        return # TODO: forward a reference-prediction pair?
    
    def report_convergence(self, *args, **kwargs) -> bool:
        """"""
        input_nodes = self.input_nodes
        assert hasattr(input_nodes[1], "output"), f"Operation {self.directory.name} cannot report convergence without forwarding."
        validator = input_nodes[1].output

        self._print(f"{validator.__class__.__name__} Convergence")
        if hasattr(validator, "report_convergence"):
            converged = validator.report_convergence()
        else:
            self._print("    >>> True  (No report available)")
            converged = True

        return converged


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import itertools

import numpy as np

from ase.geometry import find_mic

from .comparator import AbstractComparator

from ..builder.group import create_a_group


class CartesianComparator(AbstractComparator):

    dtol_avg: float = 0.1 # displacement tolerance tolerance, Ang
    dtol_std: float = 0.02 # displacement tolerance tolerance, Ang

    mic: bool = True

    group: str = None

    def __init__(self, dtol_avg=0.1, dtol_std=0.02, mic=True, group=None, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

        self.dtol_avg = dtol_avg
        self.dtol_std = dtol_std

        self.mic = mic
        self.group = group

        return
    
    def looks_like(self, a1, a2) -> bool:
        """"""
        is_similar = False
        na1, na2 = len(a1), len(a2)
        if na1 == na2:
            c1, c2 = a1.get_cell(complete=True), a2.get_cell(complete=True)
            if np.allclose(c1, c2):
                ainds = None
                if self.group is not None:
                    g1 = create_a_group(a1, self.group)
                    g2 = create_a_group(a2, self.group)
                    if g1 == g2:
                        ainds = g1
                else:
                    ainds = range(na1) # atomic indices
                if ainds is not None:
                    self._print(f"{len(ainds)}")
                    pos1, pos2 = a1.positions[ainds, :], a2.positions[ainds, :]
                    # TODO: consider permutations?
                    #perms = itertools.permutations(range(len(ainds)))
                    #for p in perms:
                    #    self._print(p)
                    #    if self.mic:
                    #        vectors, distances = find_mic(pos1 - pos2[p, :], c1, pbc=True)
                    #    else:
                    #        vectors = pos1 - pos2[p, :]
                    #    disps = np.linalg.norm(vectors, axis=1)
                    #    self._print(disps)
                    #    davg = np.average(disps) 
                    #    dstd = np.sqrt(np.var(disps))
                    #    self._print(f"davg: {davg} dstd: {dstd}")
                    if self.mic:
                        vectors, distances = find_mic(pos1 - pos2, c1, pbc=True)
                    else:
                        vectors = pos1 - pos2
                    disps = np.linalg.norm(vectors, axis=1)
                    davg = np.average(disps) 
                    dstd = np.sqrt(np.var(disps))
                    if davg <= self.dtol_avg:
                        is_similar = True
            else:
                ...
        else:
            ...

        return  is_similar


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import List
import warnings

import numpy as np
from scipy.stats import qmc

import jax
import jax.numpy as jnp
from jax import grad, value_and_grad, jit, jacfwd, jacrev

from ase import Atoms
from ase.io import read, write

from .builder import StructureModifier

"""Sample small molecule.
"""

@jit
def bond_distance(positions, indices: List[int]):
    """"""
    a1, a2 = indices
    dis = jnp.linalg.norm(positions[a2]-positions[a1])

    return dis
grad_bond_distance = grad(bond_distance, argnums=0)

@jit
def compute_bond_distances(positions: np.array, pair_indices: np.array):
    """Compute distances.

    Args:
        positions: Positions with shape of (natoms,3).
        pair_distances: Indices of atoms for distances [[0,1],[0,2]].
    
    Examples:

        >>> pairs = np.array([[0,1]])
        >>> distances = compute_bond_distances(positions, pairs)
    
    Returns:
        An array of distances.

    """
    pair_positions = jnp.take(positions, pair_indices.T, axis=0)
    dvecs = pair_positions[0] - pair_positions[1] # TODO: + pair_shifts
    distances = jnp.linalg.norm(dvecs, axis=1)

    return distances

@jit
def bond_angle(positions, indices):
    """"""
    a1, a2, a3 = indices
    dvec1 = positions[a2] - positions[a1]
    dvec2 = positions[a3] - positions[a1]

    #angle = jnp.arccos(jnp.dot(dvec1,dvec2)/jnp.linalg.norm(dvec1)/jnp.linalg.norm(dvec2))
    angle = jnp.arctan2(jnp.linalg.norm(jnp.cross(dvec1,dvec2)), jnp.dot(dvec1,dvec2))

    return angle

@jit
def compute_bond_angles(positions, trimer_indices):
    """Compute angles.

    For very small/acute angles, results by arccos are inaccurate. arctan may be 
    more effective.

    """
    trimer_positions = jnp.take(positions, trimer_indices.T, axis=0)
    # TODO: shifts
    dvecs1 = trimer_positions[1] - trimer_positions[0]
    dnorms1 = jnp.linalg.norm(dvecs1, axis=1)
    dvecs2 = trimer_positions[2] - trimer_positions[0]
    dnorms2 = jnp.linalg.norm(dvecs2, axis=1)

    angles = jnp.arccos(jnp.sum(dvecs1*dvecs2, axis=1) / dnorms1 / dnorms2)

    return angles

@jit
def pseudo_inverse_of_jacobian(jac, eps=0.0001):
    """"""
    dim = jac.shape[0]
    jac_inv = jnp.transpose(jac)@jnp.linalg.inv(jac@jnp.transpose(jac)+eps*jnp.eye(dim))

    return jac_inv


class HypercubeBuilder(StructureModifier):

    """Use hypercube sampling to generate internal coordinates.

    The implementation is not exact the same as the references. Here, we only 
    use hypercube to generate structures. No symmetry constraints is applied.

    TODO: 
        * Support dihedrals.

    References:
        [1] J. Chem. Phys. 2017, 147, 161706.
        [2] J. Chem. Phys. 2018, 149, 174114.

    """

    #: Maximum number of updates of positions.
    MAX_ATTEMPTS_UPDATE: int = 100

    #: Tolerance of target internal coordinates.
    TOL_INTCOORD: float = 1e-4

    def __init__(
        self, distances, disrange: List[float], 
        angles=None, angrange: List[float]=None, 
        substrates=None, *args, **kwargs
    ):
        """"""
        super().__init__(substrates=substrates, *args, **kwargs)

        self.dimers = np.array(distances)

        if angles:
            self.trimers = np.array(angles)
            self._use_ang = True
        else:
            self.trimers = []
            self._use_ang = False

        self.ndof = len(self.dimers) + len(self.trimers)

        bounds_ = []
        if disrange:
            if isinstance(disrange[0], list):
                bounds_.extend(disrange)
            else: # two floats
                bounds_.extend([disrange]*len(self.dimers))
        if self._use_ang:
            if angrange:
                if isinstance(angrange[0], list):
                    bounds_.extend(angrange)
                else:
                    bounds_.extend([angrange]*len(self.trimers))
            else:
                ... # TODO: set default 15,165?
        self.bounds = np.array(bounds_).T

        assert self.bounds.shape[1] == self.ndof, "Inconsistent number of freedoms and ranges."

        return
    
    def run(self, substrates: List[Atoms]=None, size: int=1, *args, **kwargs) -> List[Atoms]:
        """Modify input structures to generate random hypercube structures.
        """
        super().run(substrates=substrates, *args, **kwargs)

        frames = []
        for substrate in self.substrates:
            curr_frames = self._irun(substrate=substrate, size=size, *args, **kwargs)
            frames.extend(curr_frames)

        return frames
    
    def _irun(self, substrate: Atoms, size=1, *args, **kwargs) -> List[Atoms]:
        """"""
        # NOTE: strength new in 1.8.0
        #sampler = qmc.LatinHypercube(d=ninternals, strength=2, seed=self.rng) 
        sampler = qmc.LatinHypercube(d=self.ndof, seed=self.rng)
        samples = sampler.random(n=size)

        l_bounds, u_bounds = self.bounds
        scaled_samples = qmc.scale(samples, l_bounds, u_bounds)

        # - generate structures
        frames = []
        for sample in scaled_samples:
            curr_atoms = copy.deepcopy(substrate)
            curr_atoms = self._approx_structure(curr_atoms, sample, self.dimers, self.trimers)
            frames.append(curr_atoms)

        return frames
    
    def _approx_structure(self, atoms: Atoms, targets, pairs, trimers=None):
        """"""
        natoms = len(atoms)

        # - compute pseudo inverse of the jacobian matrix
        positions = copy.deepcopy(atoms.positions)
        for i in range(self.MAX_ATTEMPTS_UPDATE):
            distances = compute_bond_distances(positions, pairs)
            if self._use_ang:
                angles = compute_bond_angles(positions, trimers)
            else:
                angles = []
            internals = np.hstack((distances,angles))
            self._debug(f"internals: {internals}")
            disp = targets - internals 
            if np.max(np.fabs(disp)) < self.TOL_INTCOORD:
                break
            dis_jac_ = jacrev(compute_bond_distances, argnums=0)(positions, pairs)
            dis_jac = dis_jac_.reshape(-1,natoms*3)
            if self._use_ang:
                ang_jac_ = jacrev(compute_bond_angles, argnums=0)(positions, trimers)
                ang_jac = ang_jac_.reshape(-1,natoms*3)
                jac = np.vstack((dis_jac,ang_jac))
            else:
                ang_jac = []
                jac = dis_jac
            jac_inv = pseudo_inverse_of_jacobian(jac)
            positions = copy.deepcopy(positions) + jnp.reshape(jac_inv@disp, (-1,3))
        else:
            #warnings.warn("Iterative approximation is not converged.", UserWarning)
            self._print("Iterative approximation is not converged.")
        
        # - update positions
        atoms.positions = positions
        
        return atoms


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib
from typing import Union

import omegaconf

from . import registers
from ..core.variable import Variable, DummyVariable
from ..core.operation import Operation


class DescriberVariable(Variable):

    def __init__(self, directory="./", *args, **kwargs):
        """"""
        name = kwargs.pop("name", "soap")
        describer = registers.create("describer", name, convert_name=False, **kwargs)

        super().__init__(initial_value=describer, directory=directory)

        return


class describe(Operation):

    def __init__(
        self,
        structures,
        describer,
        worker=DummyVariable(),
        directory: Union[str, pathlib.Path] = "./",
    ) -> None:
        """"""
        super().__init__(
            input_nodes=[structures, describer, worker], directory=directory
        )

        return

    def _preprocess_input_nodes(self, input_nodes):
        """"""
        structures, describer, worker = input_nodes

        if isinstance(describer, dict) or isinstance(
            describer, omegaconf.dictconfig.DictConfig
        ):
            describer = DescriberVariable(
                directory=self.directory / "describer", **describer
            )

        return structures, describer, worker

    def forward(self, structures, describer, workers):
        """"""
        super().forward()

        # - verify the worker
        if workers is not None:
            nworkers = len(workers)
            assert (
                nworkers == 1
            ), f"{self.__class__.__name__} only accepts one worker but {nworkers} were given."
            worker = workers[0]
            worker.directory = self.directory/"worker"
        else:
            worker = None
        
        # - dataset?

        # - compute descriptors...
        describer.directory = self.directory
        status = describer.run(structures, worker, )

        #self.status = "finished"
        self.status = status

        return structures


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*


import pathlib


from . import AbstractPotentialManager, DummyCalculator


class EspressoManager(AbstractPotentialManager):

    name = "espresso"

    implemented_backends = ["espresso"]
    valid_combinations = (
        ("espresso", "ase")
    )

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__()

        return
    
    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        # - check backends
        super().register_calculator(calc_params, *agrs, **kwargs)

        # - parse params
        command = calc_params.get("command", "pw.x -in PREFIX.pwi > PREFIX.pwo")
        self.calc_params.update(command = command)

        pp_path = pathlib.Path(calc_params.pop("pp_path", "./")).resolve()
        if not pp_path.exists():
            raise FileNotFoundError("Pseudopotentials for espresso does not exist.")
        self.calc_params.update(pp_path=str(pp_path))

        pp_name = calc_params.get("pp_name", None)
        if pp_name is None:
            raise RuntimeError("Must set name for pseudopotentials.")

        template = calc_params.pop("template", "./espresso.pwi")
        template = pathlib.Path(template)
        if not template.exists():
            raise FileNotFoundError("Template espresso input file does not exist.")
        self.calc_params.update(template=str(template))

        from gdpx.computation.espresso import EspressoParser
        ep = EspressoParser(template=template)

        # -- kpoints that maybe read from the file
        kpts = calc_params.pop("kpts", None)
        kspacing = calc_params.pop("kspacing", None)
        koffset = calc_params.pop("koffset", 0)

        assert kpts is None or kspacing is None, "Cannot set kpts and kspacing at the same time."

        # - NOTE: check
        calc = DummyCalculator()

        if self.calc_backend == "espresso":
            from gdpx.computation.espresso import Espresso
            calc = Espresso(
                command = command,
                input_data = ep.parameters,
                pseudopotentials = pp_name,
                kspacing = kspacing,
                kpts = kpts,
                koffset = koffset
            )
        else:
            ...
        
        self.calc = calc

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*

from pathlib import Path

from . import AbstractPotentialManager


class ReaxManager(AbstractPotentialManager):

    name = "reax"
    implemented_backends = ["lammps"]

    valid_combinations = (
        ("lammps", "ase"),
        ("lammps", "lammps")
    )

    def __init__(self, *args, **kwargs):
        """"""

        return

    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)

        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", Path.cwd())

        model = calc_params.get("model", None)
        model = str(Path(model).resolve())

        if self.calc_backend == "lammps":
            from gdpx.computation.lammps import Lammps
            if model:
                pair_style = "reax/c NULL"
                pair_coeff = f"* * {model}"
                calc = Lammps(
                    command=command, directory=directory, 
                    pair_style=pair_style, pair_coeff=pair_coeff,
                    **calc_params
                )
                # - update several params
                calc.set(units="real")
                calc.set(atom_style="charge")
            else:
                calc = None
        self.calc = calc

        return


if __name__ == "__main__":
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
from typing import NoReturn, List, Union

import numpy as np
import scipy as sp
from scipy.optimize import curve_fit
from scipy.spatial import distance_matrix

from joblib import Parallel, delayed

from ase import Atoms
from ase.io import read, write

import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    #print("Used default matplotlib style.")
    ...

from ..utils.command import CustomTimer
from .validator import AbstractValidator
from .utils import wrap_traj
from ..data.array import AtomsNDArray
from ..builder.group import create_a_group

"""Measure melting point.

"""

def jpcc2020_func(T, Tm, x1, x2, x3, x4):
    """JPCC2020"""

    return x1/(1+np.exp(-x2*(T-Tm)))+x3*T+x4

def sigmoid_func(T, Tm, x1, x2, x4):
    """"""

    return x1/(1+np.exp(-x2*(T-Tm)))+x4

def get_distance_matrix(atoms: Atoms, indices=None):
    """"""
    if indices is None:
        return atoms.get_all_distances(mic=False,vector=False)
    else:
        selected_positions = atoms.positions[indices, :]
        return distance_matrix(selected_positions, selected_positions)

def _icalc_local_lindemann_index(frames, group, n_jobs=1):
    """Calculate Lindemann Index of each atom.

    Returns:
        An array with shape (natoms,)

    """
    # NOTE: Use unwrapped positions when under PBC?
    frames = wrap_traj(frames) # align structures

    group_indices = create_a_group(frames[0], group)
    natoms = len(group_indices)

    with CustomTimer("Lindemann Index"):
        distances = Parallel(n_jobs=n_jobs)(
            delayed(get_distance_matrix)(atoms, group_indices) 
            for atoms in frames
        )
    distances = np.array(distances)

    dis2 = np.square(distances)
    dis2_avg = np.average(dis2, axis=0)

    dis_avg = np.average(distances, axis=0)
    masked_dis_avg = dis_avg + np.eye(natoms) # avoid 0. in denominator
    #print(masked_dis_avg)

    q = np.sum( np.sqrt(dis2_avg - dis_avg**2) / masked_dis_avg, axis=1) / (natoms-1)

    return q


class MeltingPointValidator(AbstractValidator):

    """Estimate the melting point from a series of MD simulations.
    
    For nanoparticles, the lindeman index is used. MD simulations with various 
    initial temperatures are performed. For bulk, phase co-existence approach is used.
    Different initial temperatures are set for solid and liquid, where the steady state is
    found.

    """

    def __init__(self, group, run_fit: bool=True, start=0, temperatures: List[float]=None, fitting="sigmoid", directory: Union[str, pathlib.Path] = "./", *args, **kwargs):
        """"""
        super().__init__(directory, *args, **kwargs)

        self.group = group

        self.run_fit = run_fit

        self.start = start

        if temperatures is None:
            temperatures = []
        self.temperatures = temperatures
        
        assert fitting in ["sigmoid", "jpcc2020"], "Unsupported fitting function."
        self.fitting = fitting

        return
    
    def _process_data(self, data) -> List[List[Atoms]]:
        """"""
        data = AtomsNDArray(data)

        if data.ndim == 1:
            data = [data.tolist()]
        elif data.ndim == 2: # assume it is from extract_cache...
            data = data.tolist()
        elif data.ndim == 3: # assume it is from a compute node...
            data_ = []
            for d in data[:]: # TODO: add squeeze method?
                data_.extend(d)
            data = data_
        else:
            raise RuntimeError(f"Invalid shape {data.shape}.")

        return data

    def run(self, dataset: dict, worker=None, *args, **kwargs):
        """"""
        super().run()

        self._print("process reference ->")
        reference = dataset.get("reference")
        if reference is not None:
            reference = self._process_data(reference)
            data = self._compute_melting_point(reference, prefix="ref-")
            self._plot_figure(data[:, 1], data[:, 0], prefix="ref-", run_fit=self.run_fit)

        self._print("process prediction ->")
        prediction = dataset.get("prediction")
        if prediction is not None:
            prediction = self._process_data(prediction)
            data = self._compute_melting_point(prediction, prefix="pre-")
            self._plot_figure(data[:, 1], data[:, 0], prefix="pre-", run_fit=self.run_fit)

        return
    
    def _compute_melting_point(self, trajectories, prefix=""):
        """"""
        start, intv, end = self.start, None, None
        temperatures = self.temperatures
        assert len(trajectories) == len(temperatures), "Inconsitent number of trajectories and temperatures."

        qnames = [str(t) for t in temperatures]

        cached_data_path = self.directory / f"{prefix}qmat.txt"
        if not cached_data_path.exists():
            self._debug(f"nprocessors: {self.njobs}")
            with CustomTimer("joblib", func=self._print):
                qmat = Parallel(n_jobs=1)(
                    delayed(_icalc_local_lindemann_index)(
                        [a for a in curr_frames[start::] if a is not None], 
                        self.group, n_jobs=self.njobs
                    ) for curr_frames in trajectories
                )
            qmat = np.array(qmat)

            np.savetxt(
                cached_data_path, qmat.T,
                header=(
                    "{:>11s}"+"{:>12s}"*(len(qnames)-1)
                ).format(*qnames),
                fmt="%12.4f"
            )
        else:
            qmat = np.loadtxt(cached_data_path).T
        
        # - postprocess data
        t = temperatures
        q = np.average(qmat, axis=1)

        sorted_indices = [
            x[0] for x in sorted(enumerate(temperatures), key=lambda x: x[1])
        ]
        self._debug(sorted_indices)

        data = np.vstack(
            (
                [t[i] for i in sorted_indices],
                [q[i] for i in sorted_indices]
            )
        ).T

        np.savetxt(
            self.directory/f"{prefix}data.txt", data,
            header="{:>11s}  {:>12s}".format("temperature", "<q>"),
            fmt="%12.4f"
        )

        return data

    def _plot_figure(self, q, t: List[float], prefix="", run_fit=True):
        """"""
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))

        # - text
        fig.suptitle("Lindemann Index")

        ax.set_xlabel("Temperature [K]")
        ax.set_ylabel("$<q(T)>$")

        # - 
        ax.scatter(t, q)

        # - fitted curve
        if run_fit:
            if self.fitting == "sigmoid":
                func = sigmoid_func
                initial_guess = [np.median(t), np.max(q), 1., np.min(q)]
            elif self.fitting == "jpcc2020":
                func = jpcc2020_func
                initial_guess = [np.median(t), np.max(q), 1., 0., np.min(q)]
            coefs, cov = curve_fit(func, t, q, initial_guess, method="dogbox")
            self._debug(coefs)

            t_ = np.arange(np.min(t), np.max(t), 2.)
            q_ = func(t_, *coefs)
            ax.plot(t_, q_, label=f"$T_m={coefs[0]:>8.2f}$")

            ax.legend(loc="upper left")

        plt.savefig(self.directory/f"{prefix}mp.png")

        return
    
    def _compare_results(self):
        """Compare reference and prediction."""

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import pathlib
from typing import Optional, List, Mapping

import numpy as np
from sklearn.decomposition import PCA

from ase import Atoms
from ase.io import read, write

from ..core.register import registers
from .describer import AbstractDescriber

#try:
#    from dscribe.descriptors import SOAP
#except Exception as e:
#    print(e)

# NOTE: If there is no dscribe, this class will not be registered.
from dscribe.descriptors import SOAP


@registers.describer.register("soap")
class SoapDescriber(AbstractDescriber):

    cache_features = "features.npy"

    def __init__(self, params, *args, **kwargs) -> None:
        """"""
        super().__init__(*args, **kwargs)

        self.descriptor = copy.deepcopy(params)

        return

    def run(self, dataset, *args, **kwargs):
        """"""
        ...
        self._debug(f"n_jobs: {self.njobs}")

        # - for single system
        features = []
        for system in dataset:
            curr_frames = system._images
            if not (self.directory/system.prefix).exists():
                (self.directory/system.prefix).mkdir(parents=True)
            cache_features = self.directory/system.prefix/self.cache_features
            if not cache_features.exists():
                curr_features = self._compute_descripter(frames=curr_frames)
                np.save(cache_features, curr_features)
            else:
                curr_features = np.load(cache_features)
            features.extend(curr_features.tolist())
        features = np.array(features)
        self._debug(f"shape of features: {features.shape}")

        return features
        
    def _compute_descripter(self, frames: List[Atoms]) -> np.array:
        """Calculate vector-based descriptors.

        Each structure is represented by a vector.

        """
        self._print("start calculating features...")
        desc_params = copy.deepcopy(self.descriptor)

        soap = SOAP(**desc_params)
        ndim = soap.get_number_of_features()
        self._print(f"soap descriptor dimension: {ndim}")
        features = soap.create(frames, n_jobs=self.njobs)
        self._print("finished calculating features...")

        # - save calculated features 
        features = features.reshape(-1,ndim)

        return features


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
from typing import Any, Optional, List

from ase import Atoms
from ase.io import read, write

from ..core.node import AbstractNode


class DataSystem(AbstractNode):

    """This contains a fixed-composition system.
    """

    prefix: str = None

    _images: Optional[List[Atoms]] = None

    _tags: Optional[List[str]] = None

    def __init__(self, directory="./", pattern: str="*.xyz", *args, **kwargs) -> None:
        """"""
        super().__init__(directory=directory, *args, **kwargs)

        self.pattern = pattern
        self._process_dataset()

        return
    
    def _process_dataset(self):
        """"""
        wdir = self.directory
        self.prefix = wdir.name

        images = []
        xyzpaths = sorted(list(wdir.glob(self.pattern)))

        if self._images is None:
            self._images = []
        
        if self._tags is None:
            self._tags = []

        for p in xyzpaths:
            # -- read structures
            curr_frames = read(p, ":")
            n_curr_frames = len(curr_frames)
            self._debug(f"{p.name} nframes: {n_curr_frames}")
            self._images.extend(curr_frames)
            # -- add file prefix
            curr_tag = p.name.split(".")[0]
            self._tags.extend([curr_tag for _ in range(n_curr_frames)])

        return
    
    def get_matched_indices(self, pattern=r".*") -> List[Atoms]:
        """Get structures with given criteria.

        Args:
            origin: How these structures are created.
            generation: The model generatoin.

        """
        matched_indices = []
        for i, tag in enumerate(self._tags):
            if re.match(fr"{pattern}", tag) is not None:
                matched_indices.append(i)

        return matched_indices

    def __repr__(self) -> str:
        return f"DataSystem(nimages={len(self._images)})"


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*

""" potential wrappers
    general potential format: deepmd, eann, lasp, vasp
    dynamics backend: ase, lammps, lasp, vasp
"""


from .. import config
from ..core.register import registers


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import os
import pathlib

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.calculators.calculator import Calculator

import schnetpack as spk
import schnetpack.transform as trn
from schnetpack.data import AtomsDataModule, ASEAtomsData, load_dataset, AtomsDataFormat

from . import AbstractPotentialManager

def convert_frames(frames):
    """"""
    atoms_list, property_list = [], []
    for atoms in frames:
        propertis = {"energy": np.array([atoms.get_potential_energy()]), "forces": atoms.get_forces()}
        property_list.append(propertis)
        # set nopbc
        atoms.pbc = False
    atoms_list = frames
    #print([a.get_pbc() for a in frames])
    #assert np.all([a.get_pbc() for a in frames] == False)

    return property_list, atoms_list

def create_dataset():
    """"""
    data_paths = [
      "/users/40247882/projects/dataset/cof/H2-H2-molecule",
      "/users/40247882/projects/dataset/cof/N2-N2-molecule",
      "/users/40247882/projects/dataset/cof/O2-O2-molecule",
      "/users/40247882/projects/dataset/cof/CO-CO-molecule",
      "/users/40247882/projects/dataset/cof/H2O-H2O-molecule",
      "/users/40247882/projects/dataset/cof/H2S-H2S-molecule",
      "/users/40247882/projects/dataset/cof/CO2-CO2-molecule",
      "/users/40247882/projects/dataset/cof/NH3-H3N-molecule",
      "/users/40247882/projects/dataset/cof/CH4-CH4-molecule",
      "/users/40247882/projects/dataset/cof/methanol-CH4O-molecule",
      "/users/40247882/projects/dataset/cof/MeSCN-C2H3NS-molecule",
      "/users/40247882/projects/dataset/cof/MI-C4H6N2-molecule",
      "/users/40247882/projects/dataset/cof/benzene-C6H6-molecule",
      "/users/40247882/projects/dataset/cof/phenol-C6H6O-molecule",
      "/users/40247882/projects/dataset/cof/quinol-C6H6O2-molecule",
      "/users/40247882/projects/dataset/cof/imine-C7H7N-molecule",
      "/users/40247882/projects/dataset/cof/imine135-C9H9N3-molecule",
      "/users/40247882/projects/dataset/cof/cof5-C90H66N18O12-molecule",
      "/users/40247882/projects/dataset/cof/cof5+mimescn-C96H75N21O12S-molecule",
      "/users/40247882/projects/dataset/cof/mimescn-C6H9N3S-molecule",
    ]
    data_paths = [pathlib.Path(p) for p in data_paths]

    new_dataset = ASEAtomsData.create(
        "./dp_dataset.db",
        distance_unit="Ang",
        property_unit_dict={"energy": "eV", "forces": "eV/Ang"}
    )

    for p in data_paths:
        print(p.name)
        xyz_files = p.glob("*.xyz")
        frames = []
        for x in xyz_files:
            frames.extend(read(x, ":"))
        property_list, atoms_list = convert_frames(frames)
        new_dataset.add_systems(property_list, atoms_list)

    print('Number of reference calculations:', len(new_dataset))
    print('Available properties:')
    for p in new_dataset.available_properties:
        print('-', p)
    print()    

    example = new_dataset[0]
    print('Properties of molecule with id 0:')

    for k, v in example.items():
        print('-', k, ':', v.shape)
    
    return

def train():
    """"""
    import torch
    import torchmetrics
    import pytorch_lightning as pl

    # - create data module
    new_dataset = AtomsDataModule(
        datapath = "./new_dataset.db",
        distance_unit="Ang",
        property_units={"energy": "eV", "forces": "eV/Ang"},
        batch_size=32,
        num_train=224,
        num_val=32,
        transforms=[
            trn.ASENeighborList(cutoff=5.),
            trn.RemoveOffsets(
                "energy", remove_mean=True, remove_atomrefs=False
            ),
            trn.CastTo32()
        ],
        num_workers=1,
        pin_memory=False
    )
    new_dataset.prepare_data()
    new_dataset.setup()

    # - build model
    cutoff = 5
    n_atom_basis = 30

    # - schnet
    pairwise_distance = spk.atomistic.PairwiseDistances() # calculates pairwise distances between atoms
    radial_basis = spk.nn.GaussianRBF(n_rbf=20, cutoff=cutoff)
    #schnet = spk.representation.SchNet(
    #    n_atom_basis=n_atom_basis, n_interactions=3,
    #    radial_basis=radial_basis,
    #    cutoff_fn=spk.nn.CosineCutoff(cutoff)
    #)
    schnet = spk.representation.PaiNN(
        n_atom_basis=n_atom_basis, n_interactions=3,
        radial_basis=radial_basis,
        cutoff_fn=spk.nn.CosineCutoff(cutoff)
    )

    # - components
    pred_energy = spk.atomistic.Atomwise(n_in=n_atom_basis, output_key="energy")
    pred_forces = spk.atomistic.Forces(energy_key="energy", force_key="forces")

    nnpot = spk.model.NeuralNetworkPotential(
        representation=schnet,
        input_modules=[pairwise_distance],
        output_modules=[pred_energy, pred_forces],
        postprocessors=[
            trn.CastTo64(),
            trn.AddOffsets("energy", add_mean=True, add_atomrefs=False)
            ]
    )

    output_energy = spk.task.ModelOutput(
        name="energy",
        loss_fn=torch.nn.MSELoss(),
        loss_weight=0.01,
        metrics={
            "MAE": torchmetrics.MeanAbsoluteError()
        }
    )

    output_forces = spk.task.ModelOutput(
        name="forces",
        loss_fn=torch.nn.MSELoss(),
        loss_weight=0.99,
        metrics={
            "MAE": torchmetrics.MeanAbsoluteError()
        }
    )

    task = spk.task.AtomisticTask(
        model=nnpot,
        outputs=[output_energy, output_forces],
        optimizer_cls=torch.optim.AdamW,
        optimizer_args={"lr": 1e-4}
    )

    forcetut = './forcetut'
    if not os.path.exists(forcetut):
        os.makedirs(forcetut)

    logger = pl.loggers.TensorBoardLogger(save_dir=forcetut)
    callbacks = [
        spk.train.ModelCheckpoint(
            model_path=os.path.join(forcetut, "best_inference_model"),
            save_top_k=1,
            monitor="val_loss"
        )
    ]

    trainer = pl.Trainer(
        callbacks=callbacks,
        logger=logger,
        default_root_dir=forcetut,
        max_epochs=5, # for testing, we restrict the number of epochs
    )
    trainer.fit(task, datamodule=new_dataset)

    return

def use_model():
    """"""
    calculator = spk.interfaces.SpkCalculator(
        model_file="/mnt/scratch2/users/40247882/porous/schtrain/r0/_ensemble/m1/runs/ca9e9f60-fb47-11ed-b995-b07b25d4c022/best_model",
        neighbor_list=trn.ASENeighborList(cutoff=5.0),
        energy_key="energy",
        force_key="forces",
        stress_key="stress",
        energy_unit="eV",
        position_unit="Ang",
    )

    frames = read("/users/40247882/projects/dataset/cof/methanol-CH4O-molecule/init.xyz", ":")
    energies = np.array([a.get_potential_energy() for a in frames])
    print(energies)

    atoms = frames[0]
    atoms.calc = calculator

    #_ = atoms.get_potential_energy()
    _ = atoms.get_forces()
    #print("energy: ", atoms.get_potential_energy())
    #print("forces: ", atoms.get_forces())
    #print("stress: ", atoms.get_stress())

    print("results: ", atoms.calc.results)

    return


class SchnetManager(AbstractPotentialManager):

    name = "schnet"

    implemented_backends = ["ase"]

    valid_combinations = (
        ("ase", "ase")
    )

    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)
        self.calc = self._create_calculator(self.calc_params)

        return

    def _create_calculator(self, calc_params: dict) -> Calculator:
        """"""
        calc_params = copy.deepcopy(calc_params)

        # - some shared params
        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", pathlib.Path.cwd())

        type_list = calc_params.pop("type_list", [])
        type_map = {}
        for i, a in enumerate(type_list):
            type_map[a] = i
        
        # --- model files
        model_ = calc_params.get("model", [])
        if not isinstance(model_, list):
            model_ = [model_]

        models = []
        for m in model_:
            m = pathlib.Path(m).resolve()
            if not m.exists():
                raise FileNotFoundError(f"Cant find model file {str(m)}")
            models.append(str(m))

        # - create specific calculator
        if self.calc_backend == "ase":
            # return ase calculator
            calc = spk.interfaces.SpkCalculator(
                model_file=models[0],
                neighbor_list=trn.ASENeighborList(cutoff=5.0),
                energy_key="energy",
                force_key="forces",
                stress_key="stress",
                energy_unit="eV",
                position_unit="Ang",
            )
        elif self.calc_backend == "lammps":
            raise NotImplementedError(f"SchNet has not supported LAMMPS.")

        return calc


if __name__ == "__main__":
    #create_dataset()
    #train()
    use_model()
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np

from ase.io import read, write

from dscribe.descriptors import ValleOganov


class RmsdColvar():

    def __init__(self, data) -> None:
        """"""

        self.params = {}

        # - landmarks
        frames_ = read(data, ":")
        frames = [a[240:] for a in frames_]

        _, self.features = self.calculate_features(frames)

        return
    
    @property
    def dim(self):
        """"""

        return 1

    def calculate_features(self, frames):
        """"""
        vo2 = ValleOganov(
            species = ["Cu"],
            function = "distance",
            sigma = 10**(-5),
            n = 161,
            #n = 17,
            r_cut = 8.,
        )

        vo3 = ValleOganov(
            species = ["Cu"],
            function = "angle",
            sigma = 10**(-5),
            n = 181,
            r_cut = 4.,
        )

        #vo2_features = vo2.create(frames)
        vo2_gradients, vo2_features = vo2.derivatives(frames)
        #print(vo2_features.shape)
        #print(vo2_gradients.shape)

        return vo2_gradients, vo2_features
    
    def cvfunc(self, atoms, params):
        """"""
        gradients, features = self.calculate_features([atoms[240:]])
        features = features[np.newaxis, :] # (1, feature_dimension)
        gradients = gradients[np.newaxis, :, :, :]

        #cv = 

        return



if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from gdpx.core.variable import Variable
from gdpx.core.register import registers


@registers.variable.register
class SchedulerVariable(Variable):

    def __init__(self, directory="./", **kwargs):
        """"""
        scheduler_params = copy.deepcopy(kwargs)
        backend = scheduler_params.pop("backend", "local")
        scheduler = registers.create(
            "scheduler", backend, convert_name=True, **scheduler_params
        )
        super().__init__(initial_value=scheduler, directory=directory)

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import itertools


import numpy as np
import matplotlib.pyplot as plt

try:
    plt.style.use("presentation")
except Exception as e:
    ...

from .selector import AbstractSelector
from ..utils.geometry import wrap_traj


class BasinSelector(AbstractSelector):

    name: str = "basin"

    default_parameters: dict = dict(dispintv=0.4)  # angstrom per atom

    def __init__(self, directory="./", axis=None, *args, **kwargs) -> None:
        """"""
        super().__init__(directory, axis, *args, **kwargs)

        return

    def _mark_structures(self, data, *args, **kwargs) -> None:
        """"""
        super()._mark_structures(data, *args, **kwargs)

        print(f"data: {data}")

        if self.axis is None:
            axis = 0
        else:
            axis = self.axis

        if data.ndim >= 1:
            # assume a single trajectory
            structures = data.get_marked_structures()
            nstructures = len(structures)

            disps = self._compute_rmsd_matrix(structures)

            fig, ax = plt.subplots(1, 1, figsize=(12, 8))
            c = ax.imshow(disps, cmap="hot", interpolation="nearest")
            fig.colorbar(c)
            plt.savefig(self.directory / "rmsd.png")

            # --
            fig, ax = plt.subplots(1, 1, figsize=(12, 8))
            ax.set_title("Root Mean Squared Displacement")

            ax.scatter(range(nstructures), disps[0], c="b")
            ax.set_xlabel("number")
            ax.set_ylabel("RMSE [Å]")
            ax.tick_params(axis="y", labelcolor="b")

            energies = [a.get_potential_energy() for a in structures]
            ax2 = ax.twinx()
            ax2.scatter(range(nstructures), energies, c="r")
            ax2.set_ylabel("Potential Energy [eV]")
            ax2.tick_params(axis="y", labelcolor="r")

            plt.savefig(self.directory / "disp.png")

            # -- select
            curr_index = 0
            selected_indices = [curr_index]
            for i in range(nstructures):
                if i == curr_index:
                    for j in range(i, nstructures):
                        if np.all(disps[selected_indices, j] > self.dispintv):
                            curr_index = j
                            selected_indices.append(curr_index)
                        else:
                            ...
                else:
                    continue
            print(f"selected_indices: {selected_indices}")

            curr_markers = data.markers
            selected_markers = [curr_markers[i] for i in selected_indices]
            data.markers = selected_markers
        else:
            raise RuntimeError(
                f"Basin does not support array dimension with {data.ndim}"
            )

        return

    def _compute_rmsd_matrix(self, structures):
        """"""
        # --
        structures = wrap_traj(structures)  # TODO: copy atoms?
        nstructures, natoms = len(structures), len(structures[0])

        # --
        pairs = itertools.product(
            *[range(nstructures), range(nstructures)]
        )  # [[0, 0], [0, 1]]
        pairs = np.array(list(pairs)).T

        posmat = np.array([a.positions.flatten() for a in structures])
        disps = np.sqrt(
            np.sum(
                (np.take(posmat, pairs[0], axis=0) - np.take(posmat, pairs[1], axis=0))
                ** 2,
                axis=1,
            )
            / natoms
        )
        # print(disps.shape)
        disps = np.reshape(disps, (nstructures, nstructures))
        # print(disps.shape)

        return disps


if __name__ == "__main__":
    ...
